#!/usr/bin/perl -w

#==========================================================================
#
#    G I M S  A U T H  .  P  M
#
#    GENIPUBLIC-COPYRIGHT
#    Copyright (c) 2008-2010 University of Utah and the Flux Group.
#    All rights reserved.
#
#
#   Interface to the GENI xmlrpc interface, invoked from the web server. 
#   The certificate information is in the environment set up by apache.
#
#   $Id$
#
#==========================================================================
package GIMSAuth;


#==========================================================================
#    U S E  /  R E Q U I R E
#==========================================================================
use CGI qw(:standard);
use CGI::Carp ('fatalsToBrowser');         # Send any errors to the browser
use English;
use Frontier::Responder;
use Frontier::RPC2;
use Data::Dumper;
use POSIX;
use Crypt::X509;
use Crypt::OpenSSL::X509;

#===== Testbed libraries.
use lib '/usr/testbed/lib';
use GeniCM;
use GeniCMV2;
use Genixmlrpc;
use GeniResponse;
use libaudit;

use strict;


#==========================================================================
#    P R O T O T Y P E S
#==========================================================================
sub checkAuth();

##################### BEGIN SUBROUTINES #############################

#==========================================================================
#    C H E C K  A U T H
#
#    Returns: An error code (OEO if none) and status message.
#==========================================================================
sub checkAuth()
{

  #===== Apache does not close file descriptors before the exec, 
  #      and if this dies we are left with a giant mess.
  no warnings;
  for (my $i = 3; $i < 1024; $i++) 
  {
    POSIX:close($i);
  }

  #=====  Do this early so that we talk to the right DB.
  use vars qw($GENI_DBNAME);
  BEGIN { $GENI_DBNAME = "geni-cm"; }

  #===== Configure variables
  my $EMULAB_PEMFILE = "/usr/testbed/etc/genicm.pem";
  my $MAINSITE 	   = 0;
  my $VERSION	   = "2.0";

  #===== Geniuser.
  my $user  = "geniuser";
  my $group = "GeniSlices";

  #===== Need a command line option.
  my $debug = 0;

  my $responder; # Determined by version.

  #===== Turn off line buffering on output
  $| = 1;

  #===== Untaint the path
  $ENV{'PATH'} = '/bin:/usr/bin:/usr/local/bin';
  delete @ENV{'IFS', 'CDPATH', 'ENV', 'BASH_ENV'};

  #===== So we know who/what we are acting as.
  my $certificate = GeniCertificate->LoadFromFile($EMULAB_PEMFILE);
  if (!defined($certificate)) 
  {
    die("*** $0:\n" . "    Could not get uuid from $EMULAB_PEMFILE\n");
  }
  $ENV{'MYUUID'} = $certificate->uuid();

  #===== The URN could also come from the certificate, and that might 
  #      be preferable in some ways (if anybody is doing something 
  #      silly like authenticating with somebody else's certificate).  
  #      That would require everybody to upgrade to URNs in their 
  #      certificates, so we can't assume it yet.
  $ENV{'MYURN'} = "urn:publicid:IDN+schooner.wail.wisc.edu+authority+protogeni-test";

  #===== Make sure the client presented a valid certificate that apache 
  #      says is okay.
  #
  #      THIS HAS TO BE HERE! Recent security patches disable SSL
  #      renegotiation, which is needed when a subdir turns on ssl client
  #      verification (as httpd.conf used to). Now, we set it to "optional",
  #      which avoids the renegotiation problem, but we have to make that
  #      this interface is always invoked by a client supplying a verifiable
  #      certificate. 
  if (!(exists($ENV{'SSL_CLIENT_VERIFY'}) and  
      $ENV{'SSL_CLIENT_VERIFY'} eq "SUCCESS")) 
  {
    return(-1, "Invalid or missing certificate");
  }

  #===== In the prototype, we accept certificate signed by trusted roots
  #      (CA certs we have locally cached). This script runs as "geniuser"
  #      so that there is an emulab user context, or many of the scripts we
  #      invoke will complain and croak. 
  my $unix_uid = getpwnam("$user") or
    die("*** $0:\n".  "    No such user $user\n");
  my $unix_gid = getgrnam("$group") or
    die("*** $0:\n". "    No such group $group\n");

  #===== Flip to user and never go back
  $GID            = $unix_gid;
  $EGID           = "$unix_gid $unix_gid";
  $EUID = $UID    = $unix_uid;
  $ENV{'USER'}    = $user;
  $ENV{'LOGNAME'} = $user;

  #===== The UUID of the client certificate is in the env var 
  #      SSL_CLIENT_S_DN_CN. If it actually looks like a UUID, 
  #      then this correponds to an actual user, and the supplied 
  #      credentials/tickets must match. At present, if there is
  #      no UUID, it is another emulab making a request directly, with no user
  #      context, and we just let that pass for now.
  if (exists($ENV{'SSL_CLIENT_S_DN_CN'}) &&
      $ENV{'SSL_CLIENT_S_DN_CN'} =~ /^\w+\-\w+\-\w+\-\w+\-\w+$/) 
  {
    $ENV{'GENIUSER'} = $ENV{'SSL_CLIENT_S_DN_CN'};
    $ENV{'GENIUUID'} = $ENV{'SSL_CLIENT_S_DN_CN'};
  }
  else 
  {
    return(-1, "Invalid certificate; no UUID");
  }

  #===== The CERT data from apache holds the URN of the caller. 
  if (exists($ENV{'SSL_CLIENT_CERT'})) 
  {
    my $x509 = eval {
      Crypt::OpenSSL::X509->new_from_string($ENV{'SSL_CLIENT_CERT'}); 
    };
    if ($@) 
    {
      return(-1, "Invalid certificate: $@");
    }
    my $cert = $x509->as_string(Crypt::OpenSSL::X509::FORMAT_ASN1);

    if (!defined($cert) or $cert eq '')
    {
      return(-1, "Could not convert certificate to ASN1");
    }
    my $decoded = Crypt::X509->new( cert => $cert );
    if ($decoded->error) 
    {
      return(-1, "Error decoding certificate:" . $decoded->error);
    }

    foreach my $tmp (@{ $decoded->SubjectAltName }) 
    {
      if ($tmp =~ /^uniformResourceIdentifier=(.*)$/ or 
          $tmp =~ /^(urn:.*)$/) 
      {
        $ENV{'GENIURN'} = $1;
      }
    } # for each decoded name
  } # if the SSL_CLIENT_CERT env exists

  if (!exists($ENV{'GENIURN'}))
  {
    my $err_str = "Invalid authentication certificate; ";
    $err_str .= "no URN.  Please regenerate.";
    return(-1, $err_str);
  }


  #===== Success!
  my $return_str = "GeniUser $ENV{'GENIURN'} authenticated successfully.";
  return(0E0, $return_str);

} # end of checkAuth()


1
