#! /usr/bin/perl -w

#=============================================================================
#    G I M S  C O N T R O L  .  C G I
#
#    Test widget for GIMS project
#    Written by C. Thomas
#
#    $Id: GIMSControl.cgi,v 1.36 2010/07/12 20:28:08 cthomas Exp cthomas $
#=============================================================================


#=============================================================================
#    U S E  /  R E Q U I R E
#=============================================================================
use CGI qw(:standard);
use CGI::Carp ('fatalsToBrowser');         # Send any errors to the browser
use XML::RPC;
use XML::Simple qw(:strict);               # for converting XML <=> hash
use Data::Dumper;                          # debugging data structures
use CGI::Cookie;                           # to handle cookies
use XmlRpcUtils;                           # for writing XML/RPC code
use GIMSDBUtils;                           # to connect to database
use GIMSAuth;                              # for user authentication
use GIMSJavascript;                        # for javascript magic
use strict;


#=============================================================================
#    P R O T O T Y P E S
#=============================================================================
sub getParams($);

# Print
sub printSetParamsForm();
sub printControlForm(;$);
sub printActionSelectForm($$);
sub printConfigResultsPage(;$);
sub printExperimentStatusPage(;$);
sub printExperimentSettingsPage(;$);

# Handle
sub handleCommand($$);

# Utils
sub getDeviceCapsFromXML($);
sub sendXMLToServer($$);
sub createLibpcapString($);
sub getCookies();


#=============================================================================
#    C O N F I G
#=============================================================================
my $this_script = 'http://www.schooner.wail.wisc.edu/gims/GIMSControl.cgi';
my $agg_man_url = 'http://www.schooner.wail.wisc.edu/gims/gims_aggregate_manager.cgi';
my $log_file = 'gims_backend.log';
my $javascript_file = 'GIMSControl.js';


#=============================================================================
#    P A R A M E T E R S
#=============================================================================
my $mode = '';
my $control_action = '';
my $site_location = '';
my $device_name = '';
my $device_hostname = '';
my $device_port = '';

my @monitor_locations = ();
my $storage_type = '';
my $local_storage_dir = '';
my $s3key = '';
my $s3secret = '';
my $s3bucket = '';
my $sshuser = '';
my $sshhost = '';
my $sshport = '';
my $sshpath = '';
my $sshkey = '';
my $rollover_interval = '';
my $filter_host_addr = '';
my $host_addr = '';
my $filter_src_addr = '';
my $src_addr = '';
my $filter_dst_addr = '';
my $dst_addr = '';
my $filter_src_port = '';
my $src_port = '';
my $filter_dst_port = '';
my $dst_port = '';
my $filter_protocol = '';
my $protocol_type = '';
my $filter_bit_comp = '';
my $sampling_type = '';
my $sampling_param = '';
my $interval_secs = '';
my $aggregation_type = '';
my $anon_type = '';
my $libpcap_string = '';
my $vlan_num = '';
my $experiment_id = '';
my $meta_data_text = '';


#=============================================================================
#    G L O B A L  V A R I A B L E S
#=============================================================================
my %experiment_params = ();
my %filter_types = ();
my %storage_types = ();
my @sampling_types = ();
my @protocol_types = ();
my @aggregation_types = ();
my @anon_types = ();
my $username = '';
my $dbh = undef;
my $LOG = undef;

#=============================================================================
#    M A I N 
#=============================================================================

$| = 1; # set to unbuffered display

#===== Get a database handle
$dbh = GIMSDBUtils::getDBHandle();
die "Could not connect to database!\n" unless ref($dbh);

#==== Open log file
unless(open($LOG, ">>$log_file"))
{
  warn "Couldn't open log file! : $!\n";
}

#===== Print HTML header stuff
print "Content-type: text/html\n\n";
print qq~<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN"\n~;
print qq~"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">\n~;
print qq~<html xmlns="http://www.w3.org/1999/xhtml" lang="en-US" xml:lang="en-US">\n~;
print qq~<link href="CSS/Standard.css" rel="stylesheet" type="text/css">
\n~;
print "<head><title>GIMSControl.cgi</title></head><body>\n";
print "<body bgcolor=#83a2b8>\n";

#===== Banner
print qq~<img src='images/GIMSBanner.jpg' alt='GIMS Banner' class='banner'>\n~;

#===== Get the experiment params 
getParams(\%experiment_params);

#===== Authentication
my ($code, $status) = GIMSAuth::checkAuth();
if ($code)
{
  print "<div class=small_font>";
  print "Authentication not obtained.<br>\n";
  print "Error Code: $code<br>\n";
  print "Error Msg: $status<br>\n";
  print "</div>\n";
  #exit;
}
else
{
  print "<div class=small_font>$status</div><br>\n";
}

#===== Handle cookies
getCookies();

#===== Configure experiment mode 
if ($mode and $mode eq 'configure_experiment')
{
  # NOTE: experiment_params hash is filled by getParams() sub.

  #===== Communicate with device
  my $response = handleCommand(\%experiment_params, $agg_man_url);

  #===== Display Results 
  printConfigResultsPage($response);
} # configure_experiment 

#===== Get capture params mode
elsif ($mode and $mode eq 'get_capture_params')
{
  #===== Set up params
  my %params = ();
  $params{MethodName} = 'GetDeviceCapabilities';
  $params{ExperimentID} = $experiment_id;
  $params{SiteLocation} = $site_location;
  $params{DeviceName} = $device_name;
 
  #===== Communicate with database
  my $response = handleCommand(\%params, $agg_man_url);

  #===== Display Form
  print "<center><h1>Setup GIMS Capture Parameters</h1></center>\n";

  print $response;

  if ($response !~ m/ERROR/)
  {
    printSetParamsForm();
  }

  #===== Return to start link
  print qq~<a href="$this_script">Return to Start</a>\n~;
  
} # get_capture_params

#===== Capture control mode
elsif ($mode and $mode eq 'capture_control')
{
  my $response = '';

  #===== If we have an action already submitted, push it to the device
  if ($control_action)
  {
    #===== Write XML:RPC code
    my %control = ();
    $control{'MethodName'} = $control_action;
    $control{'ExperimentID'} = $experiment_id;

    #===== Communicate with database
    $response = handleCommand(\%control, $agg_man_url);
  } # if we're processing a previously-submitted action

  #===== Print the rest of the control form
  printControlForm($response);

} # capture_control

elsif ($mode and $mode eq 'get_status')
{
  #===== Write XML::RPC code
  my %params = ();
  $params{'MethodName'} = 'GetExperimentStatus';
  $params{'ExperimentID'} = $experiment_id;

  #===== Communicate with device
  my $response = handleCommand(\%params, $agg_man_url);

  #===== Print the results page
  printExperimentStatusPage($response);
}

elsif ($mode and $mode eq 'get_settings')
{
  my %params = ();
  $params{'MethodName'} = 'GetExperimentSettings';
  $params{'ExperimentID'} = $experiment_id;

  #===== Communicate with device
  my $response = handleCommand(\%params, $agg_man_url);

  #===== Print the results page
  printExperimentSettingsPage($response);
}

#===== Mode select
else 
{
  my %devices = ();
  my %experiments = ();

  print "<center><h1>Welcome to GIMS!</h1></center>\n";
  print "<p><p><hr>\n";

  #===== Get location/device info
  GIMSDBUtils::getDeviceInfo($dbh, \%devices);

  #===== Get experiment info
  GIMSDBUtils::getExperimentInfo($dbh, \%experiments, '', 'true');

  #===== Print the form
  printActionSelectForm(\%devices, \%experiments);
}


#==== Print HTML footer stuff
print "<p><p>\n";
print end_html, "\n";

#===== Print javascript
GIMSJavascript::printJavascript();

#===== Disconnect from database
$dbh->disconnect;

exit;

######################### BEGIN SUBROUTINES ##################################



#=============================================================================
#    G E T  P A R A M S
#=============================================================================
sub getParams($)
{
  my $config_ref = shift or
    die "Must supply hash ref to getParams()!\n";

  if ($ENV{REMOTE_USER})
  {
    $username = $ENV{REMOTE_USER};
  }
    
  if (param('mode')){              $mode = param(             'mode');}
  
  #===== Capture settings
  if (param('monitor_locations')){ @monitor_locations = param('monitor_locations');}
  if (param('storage_type')){      $storage_type = param(     'storage_type');}
  if (param('local_storage_dir')){ $local_storage_dir = param('local_storage_dir');}
  if (param('s3key')){             $s3key = param(            's3key');}
  if (param('s3secret')){          $s3secret = param(         's3secret');}
  if (param('s3bucket')){          $s3bucket = param(         's3bucket');}
  if (param('sshuser')){           $sshuser = param(          'sshuser');}
  if (param('sshhost')){           $sshhost = param(          'sshhost');}
  if (param('sshport')){           $sshport = param(          'sshport');}
  if (param('sshpath')){           $sshpath = param(          'sshpath');}
  if (param('sshkey')){            $sshkey = param(           'sshkey');}
  if (param('rollover_interval')){ $rollover_interval = param('rollover_interval');}
  if (param('filter_src_addr')){   $filter_src_addr = param(  'filter_src_addr');}
  if (param('src_addr')){          $src_addr = param(         'src_addr');}
  if (param('filter_dst_addr')){   $filter_dst_addr = param(  'filter_dst_addr');}
  if (param('dst_addr')){          $dst_addr = param(         'dst_addr');}
  if (param('filter_src_port')){   $filter_src_port = param(  'filter_src_port');}
  if (param('src_port')){          $src_port = param(         'src_port');}
  if (param('filter_dst_port')){   $filter_dst_port = param( 'filter_dst_port');}
  if (param('dst_port')){          $dst_port = param(         'dst_port');}
  if (param('filter_protocol')){   $filter_protocol = param(  'filter_protocol');}
  if (param('protocol_type')){     $protocol_type = param(    'protocol_type');}
  if (param('filter_bit_comp')){   $filter_bit_comp = param(  'filter_bit_comp');}
  if (param('sampling_type')){     $sampling_type = param(    'sampling_type');}
  if (param('sampling_param')){    $sampling_param = param(   'sampling_param');}
  if (param('interval_secs')){     $interval_secs = param(    'interval_secs');}
  if (param('aggregation_type')){  $aggregation_type = param( 'aggregation_type');}
  if (param('libpcap_string'))     {$libpcap_string = param(  'libpcap_string');}
  if (param('control_action'))     {$control_action = param(  'control_action');}
  if (param('site_location'))      {$site_location = param(   'site_location');}
  if (param('device_name'))
  {
    $device_name = param('device_name');
    if ($device_name =~ m/\:/)
    {
      ($site_location, $device_name) = split(':', $device_name);
    }
  }
  if (param('experiment_id'))     {$experiment_id = param(   'experiment_id');}
  if (param('meta_data_text'))    {$meta_data_text = param(  'meta_data_text');}
  if (param('vlan_num'))          {$vlan_num = param(        'vlan_num');}
  if (param('anon_type'))         {$anon_type = param(       'anon_type');}

  #===== Create a new libpcap string
  my %libpcap_info = ();
  $libpcap_info{'filter_src_addr'} = $filter_src_addr;
  $libpcap_info{'src_addr'} = $src_addr;
  $libpcap_info{'filter_dst_addr'} = $filter_dst_addr;
  $libpcap_info{'dst_addr'} = $dst_addr;
  $libpcap_info{'filter_src_port'} = $filter_src_port;
  $libpcap_info{'src_port'} = $src_port;
  $libpcap_info{'filter_dst_port'} = $filter_dst_port;
  $libpcap_info{'dst_port'} = $dst_port;
  $libpcap_info{'filter_protocol'} = $filter_protocol;
  $libpcap_info{'protocol_type'} = $protocol_type;
  $libpcap_info{'libpcap_string'} = $libpcap_string;
  $libpcap_string = createLibpcapString(\%libpcap_info);  

  #===== Store selected capture settings in a hash
  $config_ref->{'MethodName'} = 'ConfigureExperiment';
  $config_ref->{'ExperimentID'} = $experiment_id;
  $config_ref->{'VLAN'} = $vlan_num;
  $config_ref->{'DeviceName'} = $device_name;
  $config_ref->{'SiteLocation'} = $site_location;

  # MetaDataSpec
  $config_ref->{'MetaDataSpec'}{'UserText'} = $meta_data_text;

  # CaptureSpec
  $config_ref->{'CaptureSpec'}{'filterexpr'} = $libpcap_string;
  $config_ref->{'CaptureSpec'}{'device'} = $device_name;

  # TransformSpec
  $config_ref->{'TransformSpec'}{'SampleSpec'}{'sampletype'} = $sampling_type;
  $config_ref->{'TransformSpec'}{'SampleSpec'}{'sampleparams'}{'N'} = $sampling_param;
  $config_ref->{'TransformSpec'}{'SampleSpec'}{'sampleparams'}{'p'} = $sampling_param;

  # AnonSpec
  $config_ref->{'AnonSpec'}{'anontype'} = $anon_type;
 
  # AggregationSpec
  $config_ref->{'AggregationSpec'}{'aggtype'} = $aggregation_type;
 
  # StorageSpec
  $config_ref->{'StorageSpec'}{'storagetype'} = $storage_type;
  $config_ref->{'StorageSpec'}{'storageparams'}{'local_storage_dir'} = $local_storage_dir;
  $config_ref->{'StorageSpec'}{'storageparams'}{'s3key'} = $s3key if $s3key; 
  $config_ref->{'StorageSpec'}{'storageparams'}{'s3secret'} = $s3secret if $s3secret; 
  $config_ref->{'StorageSpec'}{'storageparams'}{'s3bucket'} = $s3bucket if $s3bucket; 
  $config_ref->{'StorageSpec'}{'storageparams'}{'sshuser'} = $sshuser if $sshuser;
  $config_ref->{'StorageSpec'}{'storageparams'}{'sshhost'} = $sshhost if $sshhost;
  $config_ref->{'StorageSpec'}{'storageparams'}{'sshport'} = $sshport if $sshport;
  $config_ref->{'StorageSpec'}{'storageparams'}{'sshpath'} = $sshpath if $sshpath;
  $config_ref->{'StorageSpec'}{'storageparams'}{'sshkey'} = $sshkey if $sshkey;
  $config_ref->{'StorageSpec'}{'storageparams'}{'rollover_interval'} = $rollover_interval if $rollover_interval;
 
  return;

} # end of getParams()


#=============================================================================
#     P R I N T  S E T  P A R A M S  F O R M
#     Build the GUI based on capabilities received by
#     getActionInfo()
#=============================================================================
sub printSetParamsForm()
{
  print qq~<form name='myForm' method="post" action="$this_script"~;  
  print "<strong>Site: $site_location</strong><br>\n";
  print "<strong>Device: $device_name</strong><br>\n";
  print "<strong>Experiment ID: $experiment_id</strong><br>\n";
  print "<table cellspacing=20>\n";
  
  #===== Storage Type 
  print "<td valign=top>\n";
  print "<strong>Storage Type:</strong><br>\n";
  foreach my $type (sort keys %storage_types)
  {
    print qq~<input type=radio name='storage_type' value='$type' ~;
    print qq~onclick='handleStorageOption();'> $type<br>~;  
  }
  print "</td>\n";

  #===== Storage Options 
  #      NOTE: This is only to show the user the options, not to 
  #      pass them to the GUI!
  print "<td valign=top>\n";
  print "<strong>Storage Options:</strong><br>\n";
  print "<input type='text' name='storage_options' value='' size=50 disabled>\n";
  print "</td>\n";

  #===== Rollover Interval
  print "<td valign=top>\n";
  print "<strong>Rollover Interval:</strong><br>\n";
  print "<select name='rollover_interval'>\n";
  print "<option value='1'>1 Minute\n";
  print "<option value='5' selected>5 Minutes\n";
  print "<option value='10'>10 Minutes\n";
  print "<option value='15'>15 Minutes\n";
  print "<option value='30'>30 Minutes\n";
  print "<option value='60'>1 Hour\n";
  print "</select>\n";
  print "</td>\n";
  print "</tr>\n";
  
  #===== Filtering
  print "<tr>\n";
  print "<td valign=top nowrap>\n";
  print "<strong>Filtering:</strong><br>\n";
  print "<table border=0>\n";
 
  if ($filter_types{'vlan_num'})
  {
    print "<tr>\n";
    print "<td align=left><input type=checkbox name='filter_vlan_num' value='true'> VLAN Num</td> \n";
    print "<td align=left><input type='text' name='vlan_num' value='' size=4></td>\n";
    print "</tr>\n";
  } 
  if ($filter_types{'src_addr'})
  {
    print "<tr>\n";
    print "<td align=left><input type=checkbox name='filter_src_addr' value='true'> Host Addr.</td> \n";
    print "<td align=left><input type='text' name='src_addr' value='144.42.12.23' size=12></td>\n";
    print "</tr>\n";
  }
  if ($filter_types{'dst_addr'})
  {
    print "<tr>\n";
    print "<td align=left><input type=checkbox name='filter_dst_addr' value='true'> Dest Addr.</td>\n";
    print "<td align=left><input type='text' name='dst_addr' value='144.42.12.1' size=12></td>\n";
    print "</tr>\n";
  }
  if ($filter_types{'src_port'})
  {
    print "<tr>\n";
    print "<td align=left><input type=checkbox name='filter_src_port' value='true'> Host Port</td>\n";
    print "<td align=left><input type='text' name='src_port' value='80' size=5></td>\n";
    print "</tr>\n";
  }
  if ($filter_types{'dst_port'})
  {
    print "<tr>\n";
    print "<td align=left><input type=checkbox name='filter_dst_port' value='true'> Dest Port</td>\n";
    print "<td align=left><input type='text' name='dst_port' value='80' size=5></td>\n";
    print "</tr>\n";
  }
  if ($filter_types{'protocol'})
  {
    print "<tr>\n";
    print "<td align=left><input type=checkbox name='filter_protocol' value='true'> Protocol</td>\n";

    #===== Protocol Type 
    print "<td valign=top>\n";
    print "<select size=1 name='protocol_type'>\n";
    foreach my $option (@protocol_types)
    {
      print "<option value='$option'> $option<br>\n";
    }
    print "</select><br>\n";
    print "</td>\n";

		print "</tr>\n";
	}
  if ($filter_types{'bit_comp'})
  {
		print "<tr>\n";
		print "<td align=left><input type=checkbox name='filter_bit_comp' value='true'> Bitwise Comp.</td>\n";
		print "<td></td>\n";
		print "</tr>\n";
  }

  print "</td>\n";
  print "</table>\n";
  print "</td>\n";

  #===== Sampling
  print "<td valign=top>\n";
  print "<strong>Sampling:</strong><br>\n";
  print "<select size=1 name='sampling_type'>\n";
  foreach my $option (@sampling_types)
  {
    print "<option value='$option'> $option<br>\n";
  }
  print "</select>\n";
  print "param: <input type='text' name='sampling_param' value='0.5' size=5>\n";
  print "</td>\n";

  #===== Aggregation
  print "<td valign=top>\n";
  print "<strong>Aggregation:</strong><br>\n";
  print "<select name='aggregation_type'>\n";
  foreach my $option (@aggregation_types)
  {
    print "<option value='$option'> $option<br>\n";
  }
  print "</select>\n";
  print "</td>\n";

  #===== Anonymization
  print "<td valign=top>\n";
  print "<strong>Anonymization:</strong><br>\n";
  print "<select name='anon_type'>\n";
  foreach my $option (@anon_types)
  {
    print "<option value='$option'> $option<br>\n";
  }
  print "</select>\n";
  print "</td>\n";

  print "</tr>\n";

  #===== Libpcap String
  print "<tr>\n";
  print "<td colspan=3 align=left>\n";
  print "<img src='images/RightArrow.gif' id='libpcap_arrow'";
  print "alt='>' width='12' height='12' align=middle/>\n";
  print "<strong>\n";
  print "<a onclick='handleLibpcapStr();'>Add Libpcap String</a><br>";
  print "<div id='libpcap_input' style='display:none'>\n";
  print "</strong>";
  print "<input type='text' name='libpcap_string' value='' size=72>\n";
  print "</div></td>\n";
  print "</tr>\n";

  #===== Meta Data
  print "<tr>\n";
  print "<td colspan=3 align=left>";
  print "<img src='images/RightArrow.gif' id='meta_data_arrow'";
  print "alt='>' width='12' height='12' align=middle/>\n";
  print "<strong>\n";
  print "<a onclick='handleMetaData();'>Add User Meta Data</a><br>";
  print "</strong>";
  print "<div id='meta_data_input' style='display:none'>\n";
  print "<textarea name='meta_data_text' rows=10 cols=72>\n";
  print "</textarea>\n";
  print "</div></td>\n";
  print "</tr>\n";
 
  #===== Buttons
  print "<tr>\n";
  print "<td colspan=3 align=center>\n";
  print "<input type='submit' value='Submit'>\n";
  print qq~&nbsp;&nbsp;&nbsp;<input type='reset' value='Reset'>\n~;
  print "</td>\n";
  print "</tr>\n";
  print "</table>\n";

  #===== Hidden fields
  print qq~<input type='hidden' name='mode' value='configure_experiment'>\n~;
  print qq~<input type="hidden" name="site_location" value="$site_location">\n~;
  print qq~<input type="hidden" name="device_name" value="$device_name">\n~;
  print qq~<input type="hidden" name="experiment_id" value="$experiment_id">\n~;
  print qq~<input type="hidden" name="local_storage_dir" value="$local_storage_dir">\n~;
  print qq~<input type="hidden" name="s3key" value="$s3key">\n~;
  print qq~<input type="hidden" name="s3secret" value="$s3secret">\n~;
  print qq~<input type="hidden" name="s3bucket" value="$s3bucket">\n~;
  print qq~<input type="hidden" name="sshuser" value="$sshuser">\n~;
  print qq~<input type="hidden" name="sshhost" value="$sshhost">\n~;
  print qq~<input type="hidden" name="sshport" value="$sshport">\n~;
  print qq~<input type="hidden" name="sshpath" value="$sshpath">\n~;
  print qq~<input type="hidden" name="sshkey" value="$sshkey">\n~;
  print qq~<input type="hidden" name="rollover_interval" value="$rollover_interval">\n~;

  print "</form>\n";
  
  return;

} # end of printSetParamsForm()


#=============================================================================
#     P R I N T  C O N T R O L  F O R M
#=============================================================================
sub printControlForm(;$)
{
  my $result = shift;

  print "<center><h1>GIMS Control Panel</h1></center>\n";
  print "<center><h2>Site: $site_location</h2></center>\n";
  print "<p><p>\n";

  #===== Experiment Details
  print "Site: $site_location<br>\n";
  print "Device: $device_name<br>\n";
  print "Experiment ID: $experiment_id<br>\n";
  print "<p><p>\n";
 
  
  print qq~<form name='myForm' method="post" action="$this_script">\n~;  

  #===== Actions Select
  print "<h2>Select an Action:</h2><p>\n";
  print "<table>\n";

  #===== Buttons
  print "<tr>\n";
  print "<td><img src='images/Start.jpeg' alt='StartExperiment' ";
  print "onClick='startExperiment();' name='start_btn'></td>\n";
  print "<td><img src='images/Pause.jpeg' alt='PauseExperiment' ";
  print "onClick='pauseExperiment();' name='pause_btn'></td>\n";
  print "<td><img src='images/Stop.jpeg' alt='StopExperiment' ";
  print "onClick='stopExperiment();' name='stop_btn'></td>\n";
  print "<td><img src='images/Status.jpeg' alt='GetExperimentStatus' ";
  print "onClick='getExperimentStatus();' name='status_btn'></td>\n";
  print "</tr>\n";

  #===== Explanations
  print "<tr>\n";
  print "<td class='med_font' align='center'>StartExperiment</td>\n";
  print "<td class='med_font' align='center'>PauseExperiment</td>\n";
  print "<td class='med_font' align='center'>StopExperiment</td>\n";
  print "<td class='med_font' align='center'>GetExperimentStatus</td>\n";
  print "</tr>\n";
  print "</table>\n";

  #===== Handle pending control actions
  if ($result)
  {
    #===== Print the results
    print "<p><p><hr>\n";
    print "Console:<p>\n";
    print "<div class='device_response2'>\n";
    print "$result";
    print "</div><p><p><hr>\n";
  }
  
  #===== Return to start link
  print qq~<a href="$this_script">Return to Start</a>\n~;

  #===== Hidden fields
  print qq~<input type='hidden' name='mode' value='capture_control'>\n~;
  print qq~<input type='hidden' name='control_action' value=''>\n~;
  print qq~<input type="hidden" name="site_location" value="$site_location">\n~;
  print qq~<input type="hidden" name="experiment_id" value="$experiment_id">\n~;

  print "</form>\n";
  
  return;
  
} # end of printControlForm()


#=============================================================================
#     P R I N T  A C T I O N  S E L E C T  F O R M
#=============================================================================
sub printActionSelectForm($$)
{
  my $devices_ref = shift or
    die "Must supply device hash ref to printActionSelectForm()!\n";
  my $exp_ref = shift or
    die "Must supply experiment hash ref to printActionSelectForm()!\n";

  print qq~<form name='myForm' method="post" action="$this_script"~;  

  print "<table cellspacing=10>\n";

  #===== Response div
  print "<div id='response_div'></div>\n";

  #===== Mode Select 
  print "<tr><td>\n";
  print "<strong>Action to perform:</strong> ";
  print "</td><td>\n";
  print "<select size=1 name='mode' onChange='handleActionSelect();'>\n";
  print "<option value='' selected> Select Action...<br>\n";
  print "<option value='create_new_experiment'> Create New Experiment<br>\n";
  print "<option value='get_capture_params'> Set Capture Parameters<br>\n";
  print "<option value='capture_control'> Control Capture Device\n";
  print "<option value='get_status'> Get Experiment Status\n";
  print "<option value='get_settings'> Get Experiment Settings\n";
  print "</select>\n";
  print "</td></tr>\n";

  #===== Device
  print "<tr><td>\n";
  print "<strong>Capture Device Name:</strong> ";
  print "</td><td>\n";
  print "<select size=1 name='device_name'>\n";
  print "<option value='' selected> Choose a device...\n";
  foreach my $location (sort keys %{$devices_ref})
  {
    foreach my $device (sort keys %{$devices_ref->{$location}})
    {
      if ($location eq $site_location and $device eq $device_name)
      {
        print "<option value='${location}:${device}' selected> ";
        print "${location}:${device}\n";
      }
      else
      {
        print "<option value='${location}:${device}'> ${location}:${device}\n";
      }
    } # for each device
  } # for each location
  print "</select>\n";
  print "</td></tr>\n";

  #===== Experiment ID 
  print "<tr><td>\n";
  print "<strong>Experiment ID:</strong> ";
  print "</td><td>\n";
  print "<div id='exp_id_div'></div>\n";
  print "</td></tr>\n";

  #===== Experiment ID Select
  print "<tr id='exp_select_row'><td>\n";
  print "<strong>Valid Experiment IDs:</strong> ";
  print "</td><td>\n";
  print "<select name='exp_select' onClick='handleExperimentIDSelect();'>\n";
  print "<option value='' selected>Select ExpID...</option>\n"; 
  foreach my $exp (sort {$exp_ref->{$b}{'last_update'} cmp
                         $exp_ref->{$a}{'last_update'}} 
                   keys %{$exp_ref})
  {
    my $status = $exp_ref->{$exp}{'status'};
    my $last_update = $exp_ref->{$exp}{'last_update'};
    print "<option value='$exp'>$exp $status ($last_update)</option>\n"; 
  }
  print "</select>\n";
  print "</td></tr>\n";

  print "</table>\n";

  #===== Button
  print "<input type='button' onClick='handleActionSubmit();' value='Submit'>\n";
   
  #===== Hidden fields
  #print qq~<input type="hidden" name="site_location" value="$site_location">\n~;
  print qq~<input type="hidden" name="experiment_id" value="">\n~;
  
  print "</form>\n";
  
  return;
  
} # end of printActionSelectForm()


#=============================================================================
#    P R I N T  C O N F I G  R E S U L T S  P A G E
#=============================================================================
sub printConfigResultsPage(;$)
{
  my $results = shift;
 
  #===== Print our header
  print "<body bgcolor=#83a2b8>\n";
  print "<center><h1>GIMS: Results</h1></center>\n";
  print "<p><p><hr>\n";

  #===== Print response string
  print $results if $results;

  #===== Return to start link
  print "<p><p>\n";
  print qq~<a href="$this_script">Return to Start</a>\n~;

  #===== Hidden fields
  print qq~<input type="hidden" name="site_location" value="$site_location">\n~;
  
  return;
  
} # end of printConfigResultsPage()


#=============================================================================
#     P R I N T  E X P E R I M E N T  S T A T U S   P A G E
#=============================================================================
sub printExperimentStatusPage(;$)
{
  my $result = shift;
 
  print "<center><h1>GIMS: Get Experiment Status</h1></center>\n";
  print "<p><p><hr>\n";
  print "<strong>Site: $site_location</strong><br>\n";
  print "<strong>Device: $device_name</strong><br>\n";
  print "<strong>Experiment ID: $experiment_id</strong><br>\n";
  print "<p><p>\n";

  print $result if $result;
  
  #===== Return to start link
  print qq~<a href="$this_script">Return to Start</a>\n~;

  return;
  
} # end of printExperimentStatusPage()


#=============================================================================
#     P R I N T  E X P E R I M E N T  S E T T I N G S   P A G E
#=============================================================================
sub printExperimentSettingsPage(;$)
{
  my $result = shift;

  print "<center><h1>GIMS: Get Experiment Settings</h1></center>\n";
  print "<p><p><hr>\n";
  print "<strong>Site: $site_location</strong><br>\n";
  print "<strong>Device: $device_name</strong><br>\n";
  print "<strong>Experiment ID: $experiment_id</strong><br>\n";
  print "<p><p>\n";

  print $result if $result;

  #===== Return to start link
  print qq~<a href="$this_script">Return to Start</a>\n~;

  return;

} # end of printExperimentSettingsPage()


#=============================================================================
#    H A N D L E  C O M M A N D
#
#    Arg1: Hash ref containing experiment settings.
#    Arg2: The aggregate manager.
#=============================================================================
sub handleCommand($$)
{
  my $exp_params_ref = shift or
    die "Must supply hash ref to handleCommand()!\n";
  my $agg_man_url = shift or
    die "Must supply aggregate manager URL to handleCommand()!\n";
  my $response = '';
  my $parse_result = '';

  #===== Generate the XML code
  my $xml_code = XmlRpcUtils::getXMLFromHash($exp_params_ref);

  # DEBUG
  #my $debug_code = $xml_code;
  #$debug_code =~ s/</&lt;/g;
  #$debug_code =~ s/>/&gt;/g;
  #print "<pre>$debug_code</pre>\n";
  
  #===== Send XML code to server
  my ($agr_mgr_data, $device_response_data) = sendXMLToServer($xml_code, 
                                                              $agg_man_url);

  #===== Get device capabilities from XML, if appropriate
  if ($device_response_data =~ m/^<\?xml version/)
  {
    $parse_result = getDeviceCapsFromXML($device_response_data);
  }

  #===== Print the response from the GIMS Aggregate Manager
  if ($agr_mgr_data or $device_response_data)
  {
    #===== Convert from XML to HTML
    $agr_mgr_data =~ s/</&lt;/g;
    $agr_mgr_data =~ s/>/&gt;/g;
  
    if ($agr_mgr_data)
    {
      $response .= "<strong>";
      $response .= "Received data from GIMS Aggregate Manager:";
      $response .= "</strong><br>\n";
      $response.= "$agr_mgr_data<p>\n";
    }
    if ($parse_result)
    {
      $response .= "<strong>";
      $response .= "Result of parsing XML to get Device Capabilities:";
      $response .= "</strong><br>\n";
      $response.= "$parse_result<p>\n";
    }
    if ($device_response_data and !$parse_result)
    {
      $device_response_data = XmlRpcUtils::getResponseString($device_response_data);
      $response .= "<strong>";
      $response .= "Received response from device:";
      $response .= "</strong><br>\n";
      $response .= "$device_response_data<p>\n";
    }
  } # if we have results to display
  else
  {
    $response .= "ERROR: Did not obtain a response from GIMS ";
    $response .= "Aggregate Manager ($agg_man_url)!\n"
  }

  return ($response);

} # end of handleCommand()


#===========================================================
#    G E T  D E V I C E  C A P S  F R O M  X M L
#
#    Arg1: The XML text to be parsed
#===========================================================
sub getDeviceCapsFromXML($)
{
  my $xml = shift or
    die("Must supply XML code to getDeviceCapsFromXML()!\n");
  my $result = '';
  my $got_params = '';
  
  # Debug    
  #my $xml_safe = $xml;
  #$xml_safe =~ s/</&lt;/g;
  #$xml_safe =~ s/>/&gt;/g;
  #$xml_safe =~ s/\n/<br>/g;
  #print "$xml_safe<p>";

  #===== Use XML::Simple to turn the XML code into a hash
  my $xml_ref = XMLin($xml, forcearray => 0, keyattr => ['name']);

  # Debug
  #$Data::Dumper::Indent = 1;
  #return(Dumper($xml_ref));
 
  #===== Check for faults
  my ($fault_code, $fault_string) = XmlRpcUtils::checkForFault($xml);
  
  #===== Handle faults
  if ($fault_string)
  {
    $result = $fault_string;
  }
  else
  {  
    #===== Get the response value from the hash
    foreach my $param_name (keys %{$xml_ref->{params}{param}{value}{struct}{member}})
    {
      my @array_contents = ();
			
      #===== Handle members that are strings
      if ($param_name eq 'device_name')
      {
        $device_name = $xml_ref->{params}{param}{value}{struct}{member}{$param_name}{value}{string};
        #$result .= "Reading device name => '$device_name'\n";
        $got_params = 1;
      }
      elsif ($param_name eq 'site_location')
      {
        $site_location = $xml_ref->{params}{param}{value}{struct}{member}{$param_name}{value}{string};
        #$result .= "Reading site_location => '$site_location'\n";
      }
      elsif ($param_name eq 'device_hostname')
      {
        $device_hostname = $xml_ref->{params}{param}{value}{struct}{member}{$param_name}{value}{string};
        #$result .= "Reading device hostname => '$device_hostname'\n";
      }		
      elsif ($param_name eq 'device_port')
      {
        $device_port = $xml_ref->{params}{param}{value}{struct}{member}{$param_name}{value}{string};
        #$result .= "Reading device port => '$device_port'\n";
      }		
			
      #===== Handle members that are arrays
      else
      {
        foreach my $value_ref (@{$xml_ref->{params}{param}{value}{struct}{member}{$param_name}{value}{array}{data}{value}})
        {
          # Pull out the array of values
          push(@array_contents, $value_ref->{string});
        }
				
        if ($param_name eq 'monitor_locations')
        {
          @monitor_locations = @array_contents;
					#$result .= "Reading monitor locations...\n";
				}
				elsif ($param_name eq 'storage_types')
				{
					#===== Store as a hash because these are mostly boolean-type values
					foreach my $type (@array_contents)
					{
						$storage_types{$type} = 1;
					}
					#$result .= "Reading storage locations...\n";
				}
				elsif ($param_name eq 'filter_types')
				{
					$result .= "Reading filtering types...\n";
					
					#===== Store as a hash because these are mostly boolean-type values
					foreach my $cap (@array_contents)
					{
						$filter_types{$cap} = 1;
					}
					#$result .= "Reading filtering types...\n";
				}
				elsif ($param_name eq 'protocol_types')
				{
					@protocol_types = @array_contents;
					#$result .= "Reading protocol types...\n";
				}
				elsif ($param_name eq 'sampling_types')
				{
					@sampling_types = @array_contents;
					#$result .= "Reading sampling types...\n";
				}
				elsif ($param_name eq 'aggregation_types')
				{
					@aggregation_types = @array_contents;
					#$result .= "Reading aggregation types...\n";
				}
				elsif ($param_name eq 'anon_types')
				{
					@anon_types = @array_contents;
					#$result .= "Reading aggregation types...\n";
				}
			} # array values
		} # for each param
		
                if ($got_params)
                {
		  $result = "Unpacked device capabilities: location='$site_location', device='$device_name'";
		} # if we got a param
	} # no faults
				
  return($result);
  
} # end of getDeviceCapsFromXML()


#=============================================================================
#    S E N D  X M L  T O  S E R V E R
#    Arg1: The XML-RPC code to send
#    Arg2: The URL of the aggregate manager
#    Returns: The response from the aggregate manager and the 
#             response from the device.
#=============================================================================
sub sendXMLToServer($$)
{
  my $xml_code = shift or
    die "Must supply XML code to sendXMLToServer()!\n";
  my $agg_man_url = shift or
    die "Must supply aggregate manager URL to sendXMLToServer()!\n";
  my $raw_xml_response = '';
  my $agr_mgr_data = '';
  my $device_response_data = '';
  
  
  #===== Open a connection to the XML:RPC server
  my $xmlrpc = XML::RPC->new($agg_man_url);

  #===== Send call to server and obtain XML-RPC response
  #      in the form of a <methodResponse> object
  $raw_xml_response = $xmlrpc->call($xml_code); 
  
  #===== Check for faults
  my ($fault_code, $fault_string) = XmlRpcUtils::checkForFault($raw_xml_response);

  if ($fault_string)
  {
    $agr_mgr_data = "Fault detected!";
    $device_response_data = $fault_string;
  }
  else
  {
		#===== Parse the XML-RPC code to a hash
		my $hr = XMLin($raw_xml_response, forcearray => 0, keyattr =>['name']);
 
    #===== Get the response values from the hash
    $agr_mgr_data = $hr->{params}{param}{value}{struct}{member}{'agr_mgr_data'}{value}{string};
    $device_response_data = $hr->{params}{param}{value}{struct}{member}{'device_response_data'}{value}{string};
  }
  
	return($agr_mgr_data, $device_response_data);

} # end of sendXMLToServer()


#===========================================================
#    W R I T E  T O  L O G
#===========================================================
sub writeToLog(;$)
{
  my $output_str = shift;

  if ($output_str)
  {
    print $LOG getTimestamp() . ": ";
    print $LOG $output_str;
  }
  
  return;
  
} # end of writeToLog()


#=============================================================================
#    C R E A T E  L I B P C A P  S T R I N G
#    Arg1: Hash ref containing setup params
#    Returns: The libpcap string
#=============================================================================
sub createLibpcapString($)
{
  my $param_ref = shift or
    die "Must supply hash ref to createLibpcapString()!\n";
  my $lib_str = '';
  
  if ($param_ref->{'filter_src_addr'})
  {
     $lib_str .= " and " if $lib_str;
     $lib_str .= "src host $param_ref->{'src_addr'}";  
    delete($param_ref->{'filter_src_addr'});
    delete($param_ref->{'src_addr'});
  }
  if ($param_ref->{'filter_dst_addr'})
  {
    $lib_str .= " and " if $lib_str;
    $lib_str .= "dst host $param_ref->{'dst_addr'}";
    delete($param_ref->{'filter_dst_addr'});
    delete($param_ref->{'dst_addr'});
  }
  if ($param_ref->{'filter_src_port'})
  {
    $lib_str .= " and " if $lib_str;
    $lib_str .= "src port $param_ref->{'src_port'}";
    delete($param_ref->{'filter_src_port'});
    delete($param_ref->{'src_port'});
  }
  if ($param_ref->{'filter_dst_port'})
  {
    $lib_str .= " and " if $lib_str;
    $lib_str .= "dst port $param_ref->{'dst_port'}";
    delete($param_ref->{'filter_dst_port'});
    delete($param_ref->{'dst_port'});
  }
  if ($param_ref->{'filter_protocol'})
  {
    $lib_str .= " and " if $lib_str;
    $lib_str .= $param_ref->{'protocol_type'};
    delete($param_ref->{'filter_protocol'});
    delete($param_ref->{'protocol_type'});
  }
  if ($param_ref->{'libpcap_string'})
  {
    $lib_str .= " and $param_ref->{'libpcap_string'}";
  }
  
  $param_ref->{'libpcap_string'} = $lib_str;
  
  return($lib_str);
  
} # end of createLibpcapString()


#=================================================================
#    G E T  C O O K I E S
#=================================================================
sub getCookies()
{
  my %cookies = ();
  
  #===== Get all cookies as a hash
  %cookies = fetch CGI::Cookie;
  
  #===== Process each cookie
  foreach my $cookie_name (keys %cookies)
  {
    if ($cookie_name eq 'device_name')
    {
      $device_name = cookie($cookie_name);
      ($site_location, $device_name) = split(/:/, $device_name);
    }
    elsif ($cookie_name eq 'experiment_id')
    {
      $experiment_id = cookie($cookie_name);
    }
  } # for each cookie
  
  return;
  
} # end of getCookies()

