#!/usr/bin/perl -w

#==============================================================================
#
#    G I M S  D B  U T I L S  .  P M 
#
#    Written by Charles Thomas (cthomas@wisc.edu)
#
#    $Id: GIMSDBUtils.pm,v 1.16 2010/07/14 19:08:54 cthomas Exp $
#==============================================================================
package GIMSDBUtils;


#=========================================================
#    U S E  R E Q U I R E
#=========================================================
use DBI;                           # for database access
use Digest::MD5;                   # to generate random experiment id
use strict;                        # always a good idea


#=========================================================
#    C O N F I G
#=========================================================
my $db_conf_file = '~/.my.cnf';


#=========================================================
#    P R O T O T Y P E S
#=========================================================
sub getDBHandle();
sub getDeviceInfo($$;$;$);
sub getExpDevices($$$;$);
sub getDeviceCapabilities($$$);
sub addExperiment($$$);
sub getExperimentInfo($$;$;$);
sub addExpDevice($$$$);
sub getExpDeviceInfo($$$;$;$);
sub updateExpDeviceSettings($$$$$);
sub getExperimentDevices($$$;$);
sub addExperimentDevice($$$$;$);
sub deleteExperimentDevice($$;$;$);
sub getExpStatus($$);
sub setExpStatus($$$);
sub getDeviceStatus($$$$);
sub setDeviceStatus($$$$$);
sub genExperimentID($);
sub getNewExpID($$);
sub expIDExists($;$);
sub expDeviceExists($$$$);
sub deviceExists($$$);
sub updateDevice($$);
sub deleteDevice($$$);


#=========================================================
#    G L O B A L  V A R I A B L E S
#=========================================================


#################### BEGIN SUBROUTINES ###################


#=========================================================
#    G E T  D B  H A N D L E
#    Returns: A database handle
#=========================================================
sub getDBHandle()
{
  my $dbh = undef;
      
  #===== Configure to talk to WiscNIC
  my @dsn = ('DBI:mysql:gims;mysql_read_default_file=' . $db_conf_file);
  
  #===== Get database handle
  $dbh = DBI->connect(@dsn);
  die "Connection to GIMS database failed!\n" unless ref($dbh);
  
  return($dbh);

} # end of getDBHandle()


#=========================================================
#    G E T  D E V I C E  I N F O 
#    Arg1: Database handle
#    Arg2: Hash ref to hold results
#    Arg3: A location (optional)
#    Arg4: A device_name (optional)
#=========================================================
sub getDeviceInfo($$;$;$)
{
  my $dbh = shift or
    die "Must supply database handle to getDeviceInfo()!\n";
  my $info_ref = shift or
    die "Must supply hash ref to getDeviceInfo()!\n";
  my $location = shift;
  my $device_name = shift;
  my $sql = undef;
  my $cursor = undef;

  #===== Prepare the SQL statement
  $sql = qq{
    SELECT 
      location,
      device_name,
      type,
      hostname,
      port,
      descr,
      capabilities,
      last_update
    FROM 
      device
  }; 
  if ($location)
  {
    $sql .= "    WHERE\n     location = '$location'";

  }
  if ($location and $device_name)
  {
    $sql .= "\n AND\n"
  }
  if ($device_name)
  {
    $sql .= "   device_name = '$device_name'";
  }

  $cursor = $dbh->prepare($sql) or 
    die "Unable to prepare database query: " . $dbh->errstr . "\n";
  $cursor->execute;

  my $LOCATION = 0;
  my $DEVICE_NAME = 1;
  my $TYPE = 2;
  my $HOSTNAME = 3;
  my $PORT = 4;
  my $DESCR = 5;
  my $CAPABILITIES = 6;
  my $LAST_UPDATE = 7;
   
  while (my @row = $cursor->fetchrow) 
  {
    #===== Initialize to prevent warnings
    foreach my $value (@row) { $value = '' unless (defined($value) and $value ne ''); }
    
    #===== Store the values
    $info_ref->{$row[$LOCATION]}{$row[$DEVICE_NAME]}{'type'} = $row[$TYPE];
    $info_ref->{$row[$LOCATION]}{$row[$DEVICE_NAME]}{'hostname'} = $row[$HOSTNAME];
    $info_ref->{$row[$LOCATION]}{$row[$DEVICE_NAME]}{'port'} = $row[$PORT];
    $info_ref->{$row[$LOCATION]}{$row[$DEVICE_NAME]}{'descr'} = $row[$DESCR];
    $info_ref->{$row[$LOCATION]}{$row[$DEVICE_NAME]}{'capabilities'} = $row[$CAPABILITIES];
    $info_ref->{$row[$LOCATION]}{$row[$DEVICE_NAME]}{'last_update'} = $row[$LAST_UPDATE];
  } # while we have more rows to read

  return;

} # end of getDeviceInfo()


#=========================================================
#    G E T  E X P  D E V I C E S
#    Arg1: Database handle
#    Arg2: Hash ref to hold results
#    Arg3: The ExperimentID 
#    Arg4: A device type (optional)
#=========================================================
sub getExpDevices($$$;$)
{
  my $dbh = shift or
    die "Must supply database handle to getExpDevices()!\n";
  my $info_ref = shift or
    die "Must supply hash ref to getExpDevices()!\n";
  my $exp_id = shift or 
    die "Must supply ExperimentID to getExpDevices()!\n";
  my $type = shift;
  my $errors = '';
  my $sql = undef;
  my $cursor = undef;

  #===== Prepare the SQL statement
  $sql = qq{
    SELECT 
      device.location,
      device.device_name,
      device.type,
      device.hostname,
      device.port,
      device.descr,
      device.capabilities,
      device.last_update
    FROM 
      device,
      exp_device
    WHERE
      exp_device.exp_id = '$exp_id' AND
      exp_device.location = device.location AND 
      exp_device.device_name = device.device_name
  }; 

  if ($type)
  {
    $sql .= " AND \n     device.type = '$type'";

  }

  $cursor = $dbh->prepare($sql) or 
    die "Unable to prepare database query: " . $dbh->errstr . "\n";
  $cursor->execute;
  $errors .= $dbh->errstr if $dbh->errstr;

  my $LOCATION = 0;
  my $DEVICE_NAME = 1;
  my $TYPE = 2;
  my $HOSTNAME = 3;
  my $PORT = 4;
  my $DESCR = 5;
  my $CAPABILITIES = 6;
  my $LAST_UPDATE = 7;
  
  while (my @row = $cursor->fetchrow) 
  {
    #===== Initialize to prevent warnings
    foreach my $value (@row) { $value = '' unless (defined($value) and $value ne ''); }

    #===== Store the values
    $info_ref->{$exp_id}{$row[$LOCATION]}{$row[$DEVICE_NAME]}{'type'} = $row[$TYPE];
    $info_ref->{$exp_id}{$row[$LOCATION]}{$row[$DEVICE_NAME]}{'hostname'} = $row[$HOSTNAME];
    $info_ref->{$exp_id}{$row[$LOCATION]}{$row[$DEVICE_NAME]}{'port'} = $row[$PORT];
    $info_ref->{$exp_id}{$row[$LOCATION]}{$row[$DEVICE_NAME]}{'descr'} = $row[$DESCR];
    $info_ref->{$exp_id}{$row[$LOCATION]}{$row[$DEVICE_NAME]}{'capabilities'} = $row[$CAPABILITIES];
    $info_ref->{$exp_id}{$row[$LOCATION]}{$row[$DEVICE_NAME]}{'last_update'} = $row[$LAST_UPDATE];
  } # while we have rows to read

  return($errors);

} # end of getExpDevices()


#=========================================================
#    G E T  D E V I C E  C A P A B I L I T I E S
#    Arg1: Database handle
#    Arg2: A location
#    Arg3: A device_name
#    Returns: Device capabilities in XML/RPC structure
#=========================================================
sub getDeviceCapabilities($$$)
{
  my $dbh = shift or
    die "Must supply database handle to getDeviceCapabilities()!\n";
  my $location = shift or
    die "Must supply site_location to getDeviceCapabilities()!\n";
  my $device_name = shift or
    die "Must supply device_name to getDeviceCapabilities()!\n";
  my $sql = undef;
  my $cursor = undef;
  my $xml = '';

  #===== Prepare the SQL statement
  $sql = qq{
    SELECT
      capabilities
    FROM 
      device
    WHERE
      location = '$location' AND
      device_name = '$device_name'
  };

  $cursor = $dbh->prepare($sql) or 
    die "Unable to prepare database query: " . $dbh->errstr . "\n";
  $cursor->execute;

  my $CAPABILITIES = 0;
  
  while (my @row = $cursor->fetchrow) 
  {
    #===== Initialize to prevent warnings
    foreach my $value (@row) { $value = '' unless (defined($value) and $value ne ''); }
    
    $xml = $row[$CAPABILITIES];
  } # while we have more rows to read

  return($xml);

} # end of getDeviceCapabilities()


#=========================================================
#    A D D  E X P E R I M E N T
#
#    Arg1: Database handle
#    Arg2: An ExperimentID
#    Arg3: A string that identifies the user
#    Returns: Errors if there are any, nothing otherwise
#=========================================================
sub addExperiment($$$)
{
  my $dbh = shift or
    die "Must supply database handle to addExperiment()!\n";
  my $exp_id = shift or
    die "Must supply hash ref to addExperiment()!\n";
  my $user = shift or
    die "Must supply user to addExperiment()!\n";
  my $sql = undef;
  my $cursor = undef;
  my $errors = '';

  #===== Prepare the SQL statement
  $sql = qq{
    INSERT INTO experiment 
    (
      exp_id,
      status,
      exp_start,
      exp_end,
      created_by,
      last_update
    )
    VALUES
    (
      '$exp_id',
      'instantiated',
      '0000-00-00 00:00:00',
      '0000-00-00 00:00:00',
      '$user',
      NULL
    )
  }; 

  $cursor = $dbh->prepare($sql) or 
    die "Unable to prepare database query: " . $dbh->errstr . "\n";
  $cursor->execute;
  $errors .= $dbh->errstr if $dbh->errstr;
  
  # DEBUG
  #$errors = $sql;

  return($errors);

} # end of addExperiment()


#=========================================================
#    G E T  E X P E R I M E N T  I N F O
#    Arg1: Database handle
#    Arg2: Hash ref to hold info
#    Arg3: An ExperimentID (optional)
#    Arg4: If true, return only active experiments (optional)
#=========================================================
sub getExperimentInfo($$;$;$)
{
  my $dbh = shift or
    die "Must supply database handle to getExperimentInfo()!\n";
  my $info_ref = shift or
    die "Must supply hash ref to getExperimentInfo()!\n";
  my $exp_id = shift;
  my $active = shift;
  my $sql = undef;
  my $cursor = undef;
  my $errors = '';

  #===== Prepare the SQL statement
  $sql = qq{
    SELECT
      exp_id,
      status,
      exp_start,
      exp_end,
      created_by,
      last_update
    FROM
      experiment
  }; 
  
  if ($exp_id and $active)
  {
    $sql .= "\nWHERE\n    exp_id='$exp_id' AND status != 'done'";
  }
  elsif ($exp_id)
  {
    $sql .= "\nWHERE\n    exp_id='$exp_id'";
  }
  elsif ($active)
  {
    $sql .= "\nWHERE\n    status != 'done'";
  }


  $cursor = $dbh->prepare($sql) or 
    die "Unable to prepare database query: " . $dbh->errstr . "\n";
  $cursor->execute;
  $errors .= $dbh->errstr if $dbh->errstr;

  my $EXPERIMENT_ID = 0;
  my $STATUS = 1;
  my $EXP_START = 2;
  my $EXP_END = 3;
  my $CREATED_BY = 4;
  my $LAST_UPDATE = 5;
  
  while (my @row = $cursor->fetchrow) 
  {
    #===== Initialize to prevent warnings
    foreach my $value (@row) { $value = '' unless (defined($value) and $value ne ''); } 

    #===== Store the values
    $info_ref->{$row[$EXPERIMENT_ID]}{'status'} = $row[$STATUS];
    $info_ref->{$row[$EXPERIMENT_ID]}{'exp_start'} = $row[$EXP_START];
    $info_ref->{$row[$EXPERIMENT_ID]}{'exp_end'} = $row[$EXP_END];
    $info_ref->{$row[$EXPERIMENT_ID]}{'created_by'} = $row[$CREATED_BY];
    $info_ref->{$row[$EXPERIMENT_ID]}{'last_update'} = $row[$LAST_UPDATE];
  } # while we have more rows to read

  return($errors);

} # end of getExperimentInfo()


#=========================================================
#    G E T  E X P  D E V I C E  I N F O
#
#    Arg1: Database handle
#    Arg2: The experiment ID
#    Arg3: The location (optional)
#    Arg4: The device name (optional)
#    Returns: Any errors.  Nothing if success.
#=========================================================
sub addExpDevice($$$$)
{
  my $dbh = shift or
    die "Must supply database handle to addExpDevice()!\n";
  my $exp_id = shift or
    die "Must supply hash ref to addExpDevice()!\n";
  my $location  = shift or
    die "Must supply location to addExpDevice()!\n";
  my $device_name = shift or
    die "Must supply device name to addExpDevice()!\n";
  my $sql = undef;
  my $cursor = undef;
  my $errors = '';

  #===== Prepare the SQL statement
  $sql = qq{
    INSERT INTO exp_device
    (
      exp_id,
      location,
      device_name,
      settings,
      last_update
    )
    VALUES
    (
      '$exp_id',
      '$location',
      '$device_name',
      NULL,
      NULL
    )
  }; 

  $cursor = $dbh->prepare($sql) or 
    die "Unable to prepare database query: " . $dbh->errstr . "\n";
  $cursor->execute;
  $errors .= $dbh->errstr if $dbh->errstr;

  return($errors);

} # end of addExpDevice()


#=========================================================
#    G E T  E X P  D E V I C E  I N F O
#
#    Arg1: Database handle
#    Arg2: The hash ref to hold results
#    Arg3: The experiment ID
#    Arg4: The location (optional)
#    Arg5: The device name (optional)
#    Returns: Any errors.  Nothing if success.
#=========================================================
sub getExpDeviceInfo($$$;$;$)
{
  my $dbh = shift or
    die "Must supply database handle to getExpDeviceInfo()!\n";
  my $info_ref = shift or
    die "Must supply hash ref to getExpDeviceInfo()!\n";
  my $exp_id = shift or
    die "Must supply exp_id to getExpDeviceInfo()!\n";
  my $location = shift;
  my $device_name = shift;
  my $sql = undef;
  my $cursor = undef;
  my $settings = '';
  my $errors = '';

  $sql = qq{
    SELECT 
      location,
      device_name,
      settings,
      status,
      last_update 
    FROM
      exp_device
    WHERE
      exp_id = '$exp_id'
  };
  
  $sql .= " AND location = '$location'" if $location;
  $sql .= " AND device_name = '$device_name" if $device_name;

  $cursor = $dbh->prepare($sql) or 
    die "Unable to prepare database query: " . $dbh->errstr . "\n";
  $cursor->execute;
  $errors .= $dbh->errstr if $dbh->errstr;

  my $LOCATION= 0;
  my $DEVICE_NAME= 1;
  my $SETTINGS= 2;
  my $STATUS = 3;
  my $LAST_UPDATE = 4;
  
  while (my @row = $cursor->fetchrow) 
  {
    #===== Initialize to prevent warnings
    foreach my $value (@row) { $value = '' unless (defined($value) and $value ne ''); } 

    #===== Store the values
    $info_ref->{$exp_id}{$row[$LOCATION]}{$row[$DEVICE_NAME]}{'settings'} = $row[$SETTINGS];
    $info_ref->{$exp_id}{$row[$LOCATION]}{$row[$DEVICE_NAME]}{'status'} = $row[$STATUS];
    $info_ref->{$exp_id}{$row[$LOCATION]}{$row[$DEVICE_NAME]}{'last_update'} = $row[$LAST_UPDATE];
  } # while we have more rows to read

  return($errors);

} # end of getExpDeviceInfo()


#=========================================================
#    U P D A T E  E X P  D E V I C E  S E T T I N G S
#    Arg1: Database handle
#    Arg2: The experiment ID
#    Arg3: The device location (e.g. 'msn')
#    Arg4: The device_name
#    Arg5: The settings used for this device/exp combination.
#    Returns: Any errors.  Nothing if success.
#=========================================================
sub updateExpDeviceSettings($$$$$)
{
  my $dbh = shift or
    die "Must supply database handle to updateExpDeviceSettings()!\n";
  my $exp_id = shift or
    die "Must supply experiment ID to updateExpDeviceSettings()!\n";
  my $location = shift or
    die "Must supply location to updateExpDeviceSettings()!\n";
  my $device_name = shift or
    die "Must supply device_name to updateExpDeviceSettings()!\n";
  my $settings = shift or
    die "Must supply settings to updateExpDeviceSettings()!\n";
  my $sql = undef;
  my $cursor = undef;
  my $errors = '';
  my $update = '';
  my %exp_settings = ();

  #===== See if we're inserting or updating
  $update = expDeviceExists($dbh, $exp_id, $location, $device_name);

  #==== Prepare the SQL statement
  if ($update)
  {
    $sql = qq{
      UPDATE
        exp_device
      SET
        settings = '$settings',
        last_update = CURRENT_TIMESTAMP() 
      WHERE
        exp_id = '$exp_id' AND
        location = '$location' AND
        device_name = '$device_name'
    };
  }
  else
  {
    $sql = qq{
      INSERT INTO exp_device
      (
        exp_id,
        location,
        device_name,
        settings,
        last_update 
      )
      values
      (
        '$exp_id',
        '$location',
        '$device_name',
        '$settings',
        CURRENT_TIMESTAMP() 
      ) 
    }; 
  } # if we're inserting

  $cursor = $dbh->prepare($sql) or 
    die "Unable to prepare database query: " . $dbh->errstr . "\n";
  $cursor->execute;
  $errors .= $dbh->errstr if $dbh->errstr;
  
  # Debug
  $errors = $sql;


  return($errors);

} # end of updateExpSettings()


#=========================================================
#    G E T  E X P E R I M E N T  D E V I C E S
#    Arg1: Database handle
#    Arg2: Hash ref to hold results
#    Arg3: The ExperimentID
#    Arg4: A type of device (optional)
#    Returns: Any errors.  Nothing if success.
#=========================================================
sub getExperimentDevices($$$;$)
{
  my $dbh = shift or
    die "Must supply database handle to getExperimentDevices()!\n";
  my $info_ref = shift or
    die "Must supply hash ref to getExperimentDevices()!\n";
  my $exp_id = shift or
    die "Must supply ExperimentID to getExperimentDevices()!\n";
  my $type = shift;
  my $sql = undef;
  my $cursor = undef;
  my $errors = '';

  #===== Prepare the SQL statement
  $sql = qq{
    SELECT
      device.location,
      device.device_name,
      device.hostname,
      device.port,
      device.descr,
      device.capabilities, 
      device.last_update
    FROM
      device,
      exp_device
    WHERE
      exp_device.exp_id = '$exp_id' AND
      exp_device.device_name = device.device_name AND
      exp_device.location = device.location
  };

  if ($type)
  {
    $sql .= " AND\n     device.type = '$type'";
  }

  $cursor = $dbh->prepare($sql) or 
    $errors .= "Unable to prepare database query: " . $dbh->errstr . "\n";

  unless ($errors)
  {
    $cursor->execute;
    $errors .= $dbh->errstr if $dbh->errstr;
  }
  unless ($errors)
  {
    my $LOCATION = 0;
    my $DEVICE_NAME = 1;
    my $HOSTNAME = 2;
    my $PORT = 3;
    my $DESCR = 4;
    my $CAPABILITIES = 5;
    my $LAST_UPDATE = 6;
  
    while (my @row = $cursor->fetchrow) 
    {
      #===== Initialize to prevent warnings
      foreach my $value (@row) {$value = '' unless (defined($value) and $value ne '');}
    
      #===== Store the values
      $info_ref->{$exp_id}{$row[$DEVICE_NAME]}{'device_name'} = $row[$DEVICE_NAME];
      $info_ref->{$exp_id}{$row[$DEVICE_NAME]}{'location'} = $row[$LOCATION];
      $info_ref->{$exp_id}{$row[$DEVICE_NAME]}{'hostname'} = $row[$HOSTNAME];
      $info_ref->{$exp_id}{$row[$DEVICE_NAME]}{'port'} = $row[$PORT];
      $info_ref->{$exp_id}{$row[$DEVICE_NAME]}{'descr'} = $row[$DESCR];
      $info_ref->{$exp_id}{$row[$DEVICE_NAME]}{'capabilities'} = $row[$CAPABILITIES];
      $info_ref->{$exp_id}{$row[$DEVICE_NAME]}{'last_update'} = $row[$LAST_UPDATE];
    } # while we have more rows to read
  } # if we have no errors

  return($errors);

} # end of getExperimentDevice()


#=========================================================
#    A D D  E X P E R I M E N T  D E V I C E
#    Arg1: A database handle
#    Arg2: The ExperimentID
#    Arg3: The site location for the device (e.g. 'msn')
#    Arg4: The device_name
#    Arg5: The device settings (XML/RPC text, optional)
#    Returns: Errors if any, nothing otherwise.
#=========================================================
sub addExperimentDevice($$$$;$)
{
  my $dbh = shift or
    die "Must supply database handle to addExperimentDevice()!\n";
  my $exp_id = shift or
    die "Must supply ExperimentID to addExperimentDevice()!\n";
  my $location = shift or 
    die "Must supply site location to addExperimentDevice()!\n";
  my $device_name = shift or
    die "Must supply device_name to addExperimentDevice()!\n";
  my $settings = shift;
  my $sql = '';
  my $cursor = '';
  my $errors = '';


  #===== Init
  $settings = '' unless $settings;

  #==== Prepare the SQL statement
  $sql = qq{
    INSERT INTO
      exp_device 
    (
      location,
      device_name,
      status,
      settings,
      last_update
    )
    VALUES
    (
      '$location',
      '$device_name,
      'instantiated',
      '$settings',
      NULL
    )
    WHERE
      exp_id = '$exp_id'
  };

  $cursor = $dbh->prepare($sql) or 
    die "Unable to prepare database query: " . $dbh->errstr . "\n";
  $cursor->execute;
  $errors .= $dbh->errstr if $dbh->errstr;
  
  # Debug
  #$errors = $sql;

  return($errors);

} # end of addExperimentDevice()


#=========================================================
#    D E L E T E  E X P E R I M E N T  D E V I C E
#
#    This sub is flexible in that if you specify a site
#    and a device, it will delete only that device, etc.
#
#    Arg1: A database handle
#    Arg2: The ExperimentID
#    Arg3: The site location for the device (e.g. 'msn', optional)
#    Arg4: The device_name (optional)
#    Returns: Any errors, nothing if success.
#=========================================================
sub deleteExperimentDevice($$;$;$)
{
  my $dbh = shift or
    die "Must supply database handle to deleteExperimentDevice()!\n";
  my $exp_id = shift or
    die "Must supply ExperimentID to deleteExperimentDevice()!\n";
  my $location = shift;
  my $device_name = shift;
  my $sql = '';
  my $cursor = '';
  my $errors = '';

  #==== Prepare the SQL statement
  $sql = qq{
    DELETE FROM
      exp_device 
    WHERE
      exp_id = '$exp_id'
  };
  if ($location and $device_name)
  {
    $sql .= " AND 
      location = '$location' AND
      device_name = '$device_name'";
  }
  elsif ($location)
  {
    $sql .= " AND
      location = '$location'";
  }

  $cursor = $dbh->prepare($sql) or 
    $errors .= "Unable to prepare database query: " . $dbh->errstr . "\n";
 
  unless ($errors)
  {
    $cursor->execute;
    $errors .= $dbh->errstr if $dbh->errstr;
  }

  return($errors);

} # end of deleteExperimentDevice()


#=========================================================
#    G E T  E X P  S T A T U S
#
#    Arg1: The database handle
#    Arg2: The ExperimentID
#=========================================================
sub getExpStatus($$)
{
  my $dbh = shift or
    die "Must supply database handle to getExpStatus()!\n";
  my $exp_id = shift or
    die "Must supply ExperimentID to getExpStatus()!\n";
  my $sql = undef;
  my $cursor = undef;
  my $status = '';

  #===== Prepare the SQL statement
  $sql = qq{
    SELECT
      status
    FROM
      experiment
    WHERE
      exp_id = '$exp_id'
  };

  $cursor = $dbh->prepare($sql) or
    die "Unable to prepare database query: " . $dbh->errstr . "\n";
  $cursor->execute;

  $status = $cursor->fetchrow;

  return($status);

} # end of getExpStatus()


#=========================================================
#    S E T  E X P  S T A T U S
#
#    Arg1: The database handle
#    Arg2: The ExperimentID
#    Arg3: The status
#=========================================================
sub setExpStatus($$$)
{
  my $dbh = shift or
    die "Must supply database handle to setExpStatus()!\n";
  my $exp_id = shift or
    die "Must supply ExperimentID to setExpStatus()!\n";
  my $status = shift or
    die "Must supply status to setExpStatus()!\n";
  my $sql = undef;
  my $cursor = undef;

  #===== Prepare the SQL statement
  $sql = qq{
    UPDATE
      experiment
    SET 
      status='$status'
    WHERE
      exp_id = '$exp_id'
  };

  $cursor = $dbh->prepare($sql) or
    die "Unable to prepare database query: " . $dbh->errstr . "\n";
  $cursor->execute;

  return;

} # end of setExpStatus()


#=========================================================
#    G E T  D E V I C E  S T A T U S
#
#    Arg1: The database handle
#    Arg2: The ExperimentID
#    Arg3: The location
#    Arg4: The device_name
#=========================================================
sub getDeviceStatus($$$$)
{
  my $dbh = shift or
    die "Must supply database handle to getDeviceStatus()!\n";
  my $exp_id = shift or
    die "Must supply ExperimentID to getDeviceStatus()!\n";
  my $location = shift or
    die "Must supply location to getDeviceStatus()!\n";
  my $device_name = shift or
    die "Must supply device_name to getDeviceStatus()!\n";
  my $sql = undef;
  my $cursor = undef;
  my $status = '';

  #===== Prepare the SQL statement
  $sql = qq{
    SELECT
      status
    FROM
      exp_device
    WHERE
      exp_id = '$exp_id' AND
      location = '$location' AND
      device_name = '$device_name'
  };

  $cursor = $dbh->prepare($sql) or
    die "Unable to prepare database query: " . $dbh->errstr . "\n";
  $cursor->execute;

  $status = $cursor->fetchrow;

  return($status);

} # end of getDeviceStatus()


#=========================================================
#    S E T  D E V I C E  S T A T U S
#
#    Arg1: The database handle
#    Arg2: The ExperimentID
#    Arg3: The location
#    Arg4: The device_name
#    Arg5: The status
#=========================================================
sub setDeviceStatus($$$$$)
{
  my $dbh = shift or
    die "Must supply database handle to setDeviceStatus()!\n";
  my $exp_id = shift or
    die "Must supply ExperimentID to setDeviceStatus()!\n";
  my $location = shift or
    die "Must supply location to setDeviceStatus()!\n";
  my $device_name = shift or
    die "Must supply device_name to setDeviceStatus()!\n";
  my $status = shift or
    die "Must supply status to setDeviceStatus()!\n";
  my $sql = undef;
  my $cursor = undef;

  #===== Prepare the SQL statement
  $sql = qq{
    UPDATE
      exp_device 
    SET 
      status='$status'
    WHERE
      exp_id = '$exp_id' AND
      location = '$location' AND
      device_name = '$device_name'
  };

  $cursor = $dbh->prepare($sql) or
    die "Unable to prepare database query: " . $dbh->errstr . "\n";
  $cursor->execute;

  return;

} # end of setDeviceStatus()


#=========================================================
#    G E N  E X P E R I M E N T  I D
#
#    Arg1: The location
#=========================================================
sub genExperimentID($)
{
  my $location = shift or
    die "Must supply location to genExperimentID()!\n";  
  my $session_id = Digest::MD5::md5_hex(Digest::MD5::md5_hex(time().{}.rand().$$));

  #===== Cut session id down to 8 characters
  $session_id = substr($session_id, 0, 8);
 
  #===== Make upper-case
  $session_id = uc($session_id);

  #===== Prepend location
  $session_id = "${location}_${session_id}";
    
  return($session_id);


} # end of genExperimentID()


#=========================================================
#    G E T  N E W  E X P  I D
#
#    Arg1: The database handle
#    Arg1: The location
#=========================================================
sub getNewExpID($$)
{
  my $dbh = shift or 
    die "Must supply database handle to getNewExpID()!\n";
  my $location = shift or
    die "Must supply location to getNewExpID()!\n";  
  my $new_exp_id = '';

  do
  {
    $new_exp_id = genExperimentID($location);
  }
  while (!$new_exp_id or expIDExists($new_exp_id));

  return($new_exp_id);

} # end of getNewExpID()


#=========================================================
#    E X P  I D  E X I S T S
#
#    Arg1: The database handle.
#    Arg2: The ExpID to test. Optional, but of little use
#          unless one is given.
#    Returns: The ExpID if it is in the database
#             nothing otherwise.
#=========================================================
sub expIDExists($;$)
{
  my $dbh = shift or 
    die "Must supply database handle to expIDExists()!\n";
  my $exp_id = shift;
  my $sql = undef;
  my $cursor = undef;
  my $exists = '';

  if ($exp_id)
  {
    #===== Prepare the SQL statement
    $sql = qq{
      SELECT 
        exp_id
      FROM 
        experiment
      WHERE
        exp_id = '$exp_id'
    };

    $cursor = $dbh->prepare($sql) or 
      die "Unable to prepare database query: " . $dbh->errstr . "\n";
    $cursor->execute;

    $exists = $cursor->fetchrow;

  } # if we have an ExpID

  return($exists);

} # end of expIDExists()


#=========================================================
#    E X P  D E V I C E  E X I S T S
#
#    Arg1: The database handle.
#    Arg2: The ExpID to test.
#    Arg3: The site location.
#    Arg4: The device_name.
#    Returns: Number of such devices (hopefully 0 or 1)
#=========================================================
sub expDeviceExists($$$$)
{
  my $dbh = shift or
    die "Must supply database handle to expDeviceExists()!\n";
  my $exp_id = shift or
    die "Must supply ExperimentID to expDeviceExists()!\n";
  my $location = shift or
    die "Must supply location to expDeviceExists()!\n";
  my $device_name = shift or
    die "Must supply device_name to expDeviceExists()!\n";
  my $sql = undef;
  my $cursor = undef;
  my $exists = '';

  #===== Prepare the SQL statement
  $sql = qq{
    SELECT 
      count(*)
    FROM 
      exp_device
    WHERE
      exp_id = '$exp_id' AND
      location = '$location' AND
      device_name = '$device_name'
  };

  $cursor = $dbh->prepare($sql) or 
    die "Unable to prepare database query: " . $dbh->errstr . "\n";
  $cursor->execute;

  $exists = $cursor->fetchrow;

  return($exists);

} # end of expDeviceExists()


#=========================================================
#    D E V I C E  E X I S T S
#
#    Arg1: The database handle.
#    Arg2: The site location.
#    Arg3: The device_name.
#    Returns: Number of such devices (hopefully 0 or 1)
#=========================================================
sub deviceExists($$$)
{
  my $dbh = shift or
    die "Must supply database handle to deviceExists()!\n";
  my $location = shift or
    die "Must supply location to deviceExists()!\n";
  my $device_name = shift or
    die "Must supply device_name to deviceExists()!\n";
  my $sql = undef;
  my $cursor = undef;
  my $exists = '';

  #===== Prepare the SQL statement
  $sql = qq{
    SELECT 
      count(*)
    FROM 
      device
    WHERE
      location = '$location' AND
      device_name = '$device_name'
  };

  $cursor = $dbh->prepare($sql) or 
    die "Unable to prepare database query: " . $dbh->errstr . "\n";
  $cursor->execute;

  $exists = $cursor->fetchrow;

  return($exists);

} # end of deviceExists()


#=========================================================
#    U P D A T E  D E V I C E
#
#    Arg1: The database handle.
#    Arg2: Hash ref containing device info.
#    Returns: Nothing for success, an err string on failure.
#=========================================================
sub updateDevice($$)
{
  my $dbh = shift or
    die "Must supply database handle to updateDevice()!\n";
  my $info_ref = shift or
    die "Must supply hash ref to updateDevice()!\n";
  my $sql = undef;
  my $cursor = undef;
  my $exists = '';
  my $error = '';
  my $status = '';

  #===== Get device info
  my $location = $info_ref->{'location'};
  my $device_name = $info_ref->{'device_name'};
  my $type = $info_ref->{'type'};
  my $hostname = $info_ref->{'hostname'};
  my $port = $info_ref->{'port'};
  my $descr = $info_ref->{'descr'};
  $descr = '' unless $descr;
  my $capabilities = $info_ref->{'capabilities'};
  $capabilities = '' unless $capabilities;


  #===== Verify
  die "Must supply location to updateDevice()!\n" unless $location;
  die "Must supply device_name to updateDevice()!\n" unless $device_name;
  die "Must supply type to updateDevice()!\n" unless $type;
  die "Must suply hostname to updateDevice()!\n" unless $hostname;
  die "Must supply port to updateDevice()!\n" unless $port;

  #===== Massage
  $descr =~ s/\'+/\'\'/g; # escape single quotes

  #===== See if we're updating or inserting
  $exists = deviceExists($dbh, $location, $device_name); 

  #===== Prepare the SQL statement
  if ($exists)
  {
    $sql = qq{
      UPDATE
        device
      SET 
        type = '$type',
        hostname = '$hostname',
        port = '$port',
        descr = '$descr',
        capabilities = '$capabilities',
        last_update = CURRENT_TIMESTAMP()
      WHERE 
        location = '$location' AND
        device_name = '$device_name'
    };
  }
  else
  {
    $sql = qq{
      INSERT INTO device
      (
        location,
        device_name,
        type,
        hostname,
        port,
        descr,
        capabilities,
        last_update
      )
      VALUES
      (
        '$location',
        '$device_name',
        '$type',
        '$hostname',
        '$port',
        '$descr',
        '$capabilities',
        CURRENT_TIMESTAMP()
      )
    };
  }

  $cursor = $dbh->prepare($sql) or 
    die "Unable to prepare database query: " . $dbh->errstr . "\n";
  $status = $cursor->execute;
  $error = $dbh->errstr if $status;
  if ($status eq '0E0')
  {
    $error = "Syntax error in SQL statement $sql!";
  }
  
  return($error);

} # end of updateDevice()


#=========================================================
#    D E L E T E  D E V I C E
#
#    Arg1: The database handle.
#    Arg2: The location
#    Arg3: The device name
#    Returns: Nothing for success, an err string on failure.
#=========================================================
sub deleteDevice($$$)
{
  my $dbh = shift or
    die "Must supply database handle to deleteDevice()!\n";
  my $location = shift or
    die "Must supply location to deleteDevice()!\n";
  my $device_name = shift or
    die "Must supply device name to deleteDevice()!\n";
  my $sql = undef;
  my $cursor = undef;
  my $error = '';
  my $status = '';

  #===== Prepare the SQL statement
  $sql = qq{
    DELETE FROM 
      device
    WHERE 
      location = '$location' AND
      device_name = '$device_name'
  };

  $cursor = $dbh->prepare($sql) or 
    die "Unable to prepare database query: " . $dbh->errstr . "\n";
  $status = $cursor->execute;
  $error = $dbh->errstr if $status;
  if ($status eq '0E0')
  {
    $error = "Device $location:$device_name was not deleted ";
    $error .= "(may not have existed)";
  }
  
  return($error);

} # end of deleteDevice() 

1
