#!/usr/bin/perl -w

#==============================================================================
#
#    X M L  R P C  U T I L S  .  P M 
#
#    Written by Charles Thomas (cthomas@wisc.edu)
#
#    $Id: XmlRpcUtils.pm,v 1.21 2010/07/14 22:11:57 cthomas Exp $
#==============================================================================
package XmlRpcUtils;


#=========================================================
#    U S E  R E Q U I R E
#=========================================================
use XML::Simple;
use Data::Dumper;
use strict;


#=========================================================
#    C O N F I G
#=========================================================
my $xml_version_str = '<?xml version="1.0" encoding="UTF-8" ?>';


#=========================================================
#    P R O T O T Y P E S
#=========================================================
sub getXMLFromHash($);
sub getCommandXML($);
sub getExtendedCommandXML($);
sub getDeviceCapsXML($);
sub getDeviceCapsHash($$);

sub writeMethodResponseXML(;$);
sub writeFaultXML(;$;$);
sub checkForFault($);

sub encodeStruct($;$);
sub encodeArray($;$);
sub encodeElement(;$;$);
sub encodeString(;$;$);
sub encodeInt(;$;$);
sub encodeDouble(;$;$);
sub encodeBoolean(;$;$);

sub getMethodCallHeaderXML($);
sub getMethodResponseHeaderXML();
sub getMethodCallFooterXML();
sub getMethodResponseFooterXML();
sub getValueFromHash($$);
sub getFormattedTextFromXML($);
sub getKeyValuePairs($);

sub getExperimentID($);
sub getCommand($);
sub getLocationAndDeviceName($);
sub getResponseString($);
 
sub getHttpHeader(;$);


#=========================================================
#    G L O B A L  V A R I A B L E S
#=========================================================


#################### BEGIN SUBROUTINES ###################


#=========================================================
#    G E T  X M L  F R O M  H A S H
#
#    Arg1: Hash containing method call information.
#    Returns: Formatted XML code as a scalar.
#=========================================================
sub getXMLFromHash($)
{
  my $params_ref = shift or 
    die "Must supply hash ref to getXMLFromHash()!\n";
  my $xml_code = '';
  
  my $method_name = $params_ref->{'MethodName'};
  
  #===== Parse the method name
  if ($method_name eq 'StartExperiment' or
      $method_name eq 'StopExperiment' or
      $method_name eq 'PauseExperiment' or
      $method_name eq 'GetExperimentStatus' or
      $method_name eq 'GetExperimentSettings')
  {
    $xml_code .= getCommandXML($params_ref);
  }
  elsif ($method_name eq 'ConfigureExperiment' or
         $method_name eq 'GetDeviceCapabilities')
  {
    $xml_code .= getExtendedCommandXML($params_ref);
  }
  else
  {
    $xml_code .= writeFaultXML(-1, "Unknown command $method_name in getXMLFromHash()!");
  }

  return ($xml_code);
  
} # end of getXMLFromHash()


#=========================================================
#    G E T  C O M M A N D  X M L
#
#    Arg1: Hash containing method call information.
#    Returns: Formatted XML code as a scalar.
#=========================================================
sub getCommandXML($)
{
  my $params_ref = shift or 
    die "Must supply hash ref to getCommandXML()!\n";
  my $xml_code = '';
  
  my $method_name = $params_ref->{'MethodName'};
  my $exp_id = $params_ref->{'ExperimentID'};
  
  #===== Write code if we have what we need
  if (defined($exp_id) and $method_name)
  {
    $xml_code .= getMethodCallHeaderXML($method_name);
    $xml_code .= "<params>\n";
    $xml_code .= "  <param>\n";
    $xml_code .= "    <value>\n";
    $xml_code .= "      <struct>\n";
    $xml_code .= "        <member>\n";
    $xml_code .= "          <name>ExperimentID</name>\n";
    $xml_code .= "          <value>" . encodeString($exp_id) . "</value>\n";
    $xml_code .= "        </member>\n";
    $xml_code .= "      </struct>\n";
    $xml_code .= "    </value>\n";
    $xml_code .= "  </param>\n";
    $xml_code .= "</params>\n";
    $xml_code .= getMethodCallFooterXML();
  }
  else
  {
    #===== Handle error
    $xml_code = writeFaultXML(-1, "Missing parameter to getCommandXML()!");
  }

  return ($xml_code);

} # end of getCommandXML()


#=========================================================
#    G E T  E X T E N D E D  C O M M A N D  X M L
#
#    Arg1: Hash containing method call information.
#    Returns: Formatted XML code as a scalar.
#=========================================================
sub getExtendedCommandXML($)
{
  my $params_ref = shift or 
    die "Must supply hash ref to getExtendedCommandXML()!\n";
  my $xml_code = '';

  my $method_name = $params_ref->{'MethodName'};
  my $exp_id = $params_ref->{'ExperimentID'};

  #===== Write code if we have what we need
  if (defined($exp_id) and $method_name)
  {
    $xml_code .= getMethodCallHeaderXML($method_name);
    $xml_code .= "<params>\n";
    $xml_code .= "  <param>\n";
    $xml_code .= "    <value>\n";
    $xml_code .= encodeStruct($params_ref, 4);
    $xml_code .= "    </value>\n";
    $xml_code .= "  </param>\n";
    $xml_code .= "</params>\n";
    $xml_code .= getMethodCallFooterXML();
  }
  else
  {
    #===== Handle error
    $xml_code = writeFaultXML(-1, "Missing parameter to getExtendedCommandXML()!");
  }


  return ($xml_code);

} # end of geExtendedCommandXML()


#=========================================================
#    G E T  D E V I C E  C A P S  X M L
#
#    Arg1: Hash containing information.
#    Returns: Formatted XML code as a scalar.
#=========================================================
sub getDeviceCapsXML($)
{
  my $params_ref = shift or 
    die "Must supply hash ref to getDeviceCapsXML()!\n";
  my $xml_code = '';

  $xml_code .= getMethodResponseHeaderXML(); 
  $xml_code .= "<params>\n";
  $xml_code .= "  <param>\n";
  $xml_code .= "    <value>\n";
  $xml_code .= encodeStruct($params_ref, 4);
  $xml_code .= "    </value>\n";
  $xml_code .= "  </param>\n";
  $xml_code .= "</params>\n";
  $xml_code .= getMethodResponseFooterXML();

  return($xml_code);

} # end of getDeviceCapsXML($)


#=========================================================
#    G E T  D E V I C E  C A P S  H A S H
#
#    Arg1: Scalar containing information as XML/RPC.
#    Returns: Hash ref 
#=========================================================
sub getDeviceCapsHash($$)
{
  my $xml_code = shift or 
    die "Must supply XML/RPC code to getDeviceCapsHash()!\n";
  my $caps_ref = shift or
    die "Must supply hash ref to getDeviceCapsHash()!\n";
  my @lines = split('\n', $xml_code);
  my $key = '';
  my $value = '';
 
  #===== Process each line 
  foreach my $line (@lines)
  {
    if ($line =~ m/\<name\>(.+)\<\/name\>/)
    {
      $key = $1;
    }
    elsif ($line =~ m/\<value\>\<string\>(.+)\<\/string\>\<\/value\>/)
    {
      $value = $1;
      $caps_ref->{$key}{$value} = 1;
    }
  }

  return;

} # end of getDeviceCapsHash()


#=========================================================
#    W R I T E  M E T H O D  R E S P O N S E  X M L
#
#    Arg1: Scalar containing response
#    Returns: Formatted XML code as a scalar.
#=========================================================
sub writeMethodResponseXML(;$)
{
  my $response = shift;
  my $xml_code = '';
  
  $xml_code .= getMethodResponseHeaderXML();
  $xml_code .= "<params>\n";
  $xml_code .= "  <param>\n";
  $xml_code .= "    <value><string>$response</string></value>\n";
  $xml_code .= "  </param>\n";
  $xml_code .= "</params>\n";
  $xml_code .= getMethodResponseFooterXML();  
  
  return($xml_code);
  
} # end of writeMethodResponseXML()


#=========================================================
#    W R I T E  F A U L T  X M L
#
#    Arg1: Scalar containing numeric fault code.
#    Arg2: Scalar containing fault string.
#    Returns: Formatted XML code as a scalar.
#=========================================================
sub writeFaultXML(;$;$)
{
  my $fault_code = shift;
  my $fault_string = shift;
  my $xml_code = '';
  
  $xml_code .= getMethodResponseHeaderXML();
  
  $xml_code .= "<fault>\n";
  $xml_code .= "  <value>\n";
  $xml_code .= "    <struct>\n";
  
  $xml_code .= "    <member>\n";
  $xml_code .= "    <name>faultCode</name>\n";
  $xml_code .= "    <value>" . encodeInt($fault_code) . "</value>\n";
  $xml_code .= "    </member>\n";
  
  $xml_code .= "    <member>\n";
  $xml_code .= "    <name>faultString</name>\n";
  $xml_code .= "    <value>" . encodeString($fault_string) . "</value>\n";
  $xml_code .= "    </member>\n";

  $xml_code .= "    </struct>\n";
  $xml_code .= "  </value>\n";
  $xml_code .= "</fault>\n";
  
  $xml_code .= getMethodResponseFooterXML();  
  
  return($xml_code);
  
} # end of writeFaultXML()


#=============================================================================
#    C H E C K  F O R  F A U L T
#    Arg1: The XML-RPC code to parse
#    Returns: (faultCode, faultString) if there was a fault.
#             undef, undef otherwise
#=============================================================================
sub checkForFault($)
{
  my $xml_code = shift or
    die "Must supply XML code to checkForFault()!\n";
  my $fault_code = undef;
  my $fault_string = undef;
  
  #===== Parse the XML-RPC code to a hash
  my $hr = XMLin($xml_code, forcearray => 0, keyattr =>['name']);
  
  if (ref($hr->{fault}))
  {
    $fault_string = "ERROR: ";
    foreach my $param_name (keys %{$hr->{fault}{value}{struct}{member}})
    {
      if ($param_name eq 'faultCode')
      {
        $fault_code = $hr->{fault}{value}{struct}{member}{$param_name}{value}{string};
      }               
      elsif ($param_name eq 'faultString')
      {
        $fault_string .= $hr->{fault}{value}{struct}{member}{$param_name}{value}{string};
      }               
    } # for each param 
  } # if it's a fault
  
  return($fault_code, $fault_string);
  
} # end of checkForFault()


#=========================================================
#    E N C O D E  S T R U C T
#
#    Arg1: Hash ref containing the data to be encoded.
#    Arg2: Level of indentation to start at (optional).
#    Returns: Formatted XML code as a scalar.
#=========================================================
sub encodeStruct($;$)
{
  my $hash_ref = shift or
    die "Must supply hash ref to encodeStruct()!\n";
  my $level = shift;
  my $xml_code = '';
  my $pad = '';

  #===== Create padding
  $level = 0 unless $level;
  foreach (my $i = 0; $i < $level; $i++){$pad .= '  ';}
  my $pad2 = "$pad  ";
  my $pad4 = "$pad2  ";
  
  #===== Start struct
  $xml_code .= "$pad<struct>\n";
  
  #===== Handle each element of the hash
  while (my ($key, $value) = each %{$hash_ref})
  {    
    $xml_code .= "$pad2<member>\n";
    $xml_code .= "$pad4<name>$key</name>\n";
    
    if (ref($value) eq 'HASH')
    {
      $xml_code .= "$pad4<value>\n";
      $xml_code .= encodeStruct($value, $level+3);
      $xml_code .= "$pad4</value>\n";
    }
    elsif (ref($value) eq 'ARRAY')
    {
      $xml_code .= "$pad4<value>\n";
      $xml_code .= encodeArray($value, $level+3);
      $xml_code .= "$pad4</value>\n";
    }
    elsif (ref($value) eq 'SCALAR')
    {
      $xml_code .= "$pad4<value>";
      $xml_code .= encodeElement($value, $level+1);
      $xml_code .= "</value>\n";
    }
    elsif (!ref($value) and $value)
    {
      $xml_code .= "$pad4<value>";
      $xml_code .= encodeElement($value, $level+1); 
      $xml_code .= "</value>\n";
    }
    else
    {
      #===== Blank
      $xml_code .= "$pad4<value><string></string></value>\n";
    }

    $xml_code .= "$pad2</member>\n";

  } # for each key

  $xml_code .= "$pad</struct>\n";
     
  return($xml_code);
  
} # end of encodeStruct()


#=========================================================
#    E N C O D E  A R R A Y
#
#    Arg1: Hash ref containing info to be encoded.
#    Arg2: Level of indentation to start at (optional).
#    Returns: Formatted XML code as a scalar.
#=========================================================
sub encodeArray($;$)
{
  my $array_ref = shift or
    die "Must array ref to encodeArray()!\n";
  my $level = shift;
  my $xml_code = '';
  my $pad = '';

  #===== Create padding
  $level = 0 unless $level;
  foreach (my $i = 0; $i < $level; $i++){$pad .= '  ';}
  my $pad2 = "$pad  ";

  $xml_code .= "$pad<array><data>\n";
  
  #===== Go through each element
  foreach my $element (@$array_ref)
  {
    $xml_code .= "$pad2<value>";
    $xml_code .= encodeElement($element, $level+1);
    $xml_code .= "</value>\n";
  }
  
  $xml_code .= "$pad</data></array>\n";

  return($xml_code);

} # end of encodeArray()


#=========================================================
#    E N C O D E  E L E M E N T
#
#    Examines the element and decides if it's an int,
#    double or string.  Then calls the appropriate
#    sub to process it.
#
#    Arg1: Element (int, double, string) to be encoded.
#    Arg2: Level of indentation to start at (optional).
#    Returns: Formatted XML code as a scalar.
#=========================================================
sub encodeElement(;$;$)
{
  my $element = shift;
  my $level = shift;
  my $xml_code = '';
  
  if ($element =~ m/^\-*\d+$/) # number
  {
    $xml_code .= encodeInt($element, $level); 
  }
  elsif ($element =~ m/^\-*\d*\.{1}\d*$/) # double
  {
    $xml_code .= encodeDouble($element, $level); 
  }
  else # string
  {
    $xml_code .= encodeString($element, $level); 
  }
  
  return($xml_code);

} # end of encodeElement()


#=========================================================
#    E N C O D E  S T R I N G
#
#    Arg1: String to be encoded.
#    Arg2: Level of indentation to start at (optional).
#    Returns: Formatted XML code as a scalar.
#=========================================================
sub encodeString(;$;$)
{
  my $str = shift;
  my $level = shift;
  my $xml_code = '';
  my $pad = '';

  #===== Create padding
  $level = 0 unless $level;
  foreach (my $i = 0; $i < $level; $i++){$pad .= '  ';}

  $xml_code .= "<string>$str</string>";

  return($xml_code);

} # end of encodeString()


#=========================================================
#    E N C O D E  I N T
#
#    Arg1: Int to be encoded.
#    Arg2: Level of indentation to start at (optional).
#    Returns: Formatted XML code as a scalar.
#=========================================================
sub encodeInt(;$;$)
{
  my $int = shift;
  my $level = shift;
  my $xml_code = '';
  my $pad = '';

  #===== Create padding
  $level = 0 unless $level;
  foreach (my $i = 0; $i < $level; $i++){$pad .= '  ';}
  
  $xml_code .= "<int>$int</int>";

  return($xml_code);
  
} # end of encodeInt()


#=========================================================
#    E N C O D E  D O U B L E
#
#    Arg1: Double (float) to be encoded.
#    Arg2: Level of indentation to start at (optional).
#    Returns: Formatted XML code as a scalar.
#=========================================================
sub encodeDouble(;$;$)
{
  my $double = shift;
  my $level = shift;
  my $xml_code = '';
  my $pad = '';

  #===== Create padding
  $level = 0 unless $level;
  foreach (my $i = 0; $i < $level; $i++){$pad .= '  ';}
  
  $xml_code .= "<double>$double</double>";

  return($xml_code);
  
} # end of encodeDouble()


#=========================================================
#    E N C O D E  B O O L E A N
#
#    Arg1: Boolean [0|1] to be encoded.
#    Arg2: Level of indentation to start at (optional).
#    Returns: Formatted XML code as a scalar.
#=========================================================
sub encodeBoolean(;$;$)
{
  my $bool = shift;
  my $level = shift;
  my $xml_code = '';
  my $pad = '';

  #===== Create padding
  $level = 0 unless $level;
  foreach (my $i = 0; $i < $level; $i++){$pad .= '  ';}
  
  $xml_code .= "<boolean>$bool</boolean>";

  return($xml_code);
  
} # end of encodeBoolean()


#=========================================================
#    G E T  M E T H O D  C A L L  H E A D E R  X M L
#
#    Arg1: Method name
#    Returns: Formatted XML code as a scalar.
#=========================================================
sub getMethodCallHeaderXML($)
{
  my $method_name = shift or
    die "Must supply method name to getMethodCallHeaderXML()!\n";
  my $xml_code = '';

  $xml_code .=  qq~$xml_version_str\n~;
  $xml_code .=  qq~<methodCall>\n~;
  $xml_code .=  qq~<methodName>$method_name</methodName>\n~;

  return($xml_code);

} # end of getMethodCallHeaderXML()


#=============================================================
#    G E T  M E T H O D  R E S P O N S E  H E A D E R  X M L
#
#    Returns: Formatted XML code as a scalar.
#==============================================================
sub getMethodResponseHeaderXML()
{
  my $xml_code = '';

  $xml_code .=  qq~$xml_version_str\n~;
  $xml_code .=  qq~<methodResponse>\n~;

  return($xml_code);

} # end of getMethodResponseHeaderXML()


#=========================================================
#    G E T  M E T H O D  C A L L  F O O T E R  X M L
#
#    Returns: Formatted XML code as a scalar.
#=========================================================
sub getMethodCallFooterXML()
{
  my $xml_code = '';

  $xml_code .=  qq~</methodCall>\n~;

  return($xml_code);

} # end of getMethodCallFooterXML()


#=============================================================
#    G E T  M E T H O D  R E S P O N S E  F O O T E R  X M L
#
#    Returns: Formatted XML code as a scalar.
#=============================================================
sub getMethodResponseFooterXML()
{
  my $xml_code = '';

  $xml_code .=  qq~</methodResponse>\n~;

  return($xml_code);

} # end of getMethodResponseFooterXML()


#=============================================================
#    G E T  V A L U E  F R O M  H A S H
#
#    Arg1: Reference to the hash to be searched.
#    Arg2: The key we're looking for.
#    Returns: The value if the key was found, undef otherwise.
#=============================================================
sub getValueFromHash($$)
{
  my $info_ref = shift or
    die "Must supply hash ref to getValueFromHash()!\n";
  my $search_key = shift or
    die "Must supply key to getValueFromHash()!\n";
  my $value = undef;

  #===== Make sure it's a ref
  die "Not a hash ref at getValueFromHash()!\n" unless ref($info_ref);

  foreach my $key (keys %{$info_ref})
  {
    if ($key eq $search_key)
    {
      #===== We found our key, get the XML value
      $value = $info_ref->{$key}{value}{string};
      last;
    }
    elsif (ref($info_ref->{$key}) eq 'HASH')
    {
      $value = getValueFromHash($info_ref->{$key}, $search_key);
      last if $value;
    } # if the key is a ref itself
  } # foreach key

  return($value);

} # end of getValueFromHash()


#=============================================================
#    G E T  F O R M A T T E D  T E X T  F R O M  X M L
#
#    Arg1: The XML/RPC code to be transformed.
#    Returns: Formatted text.
#=============================================================
sub getFormattedTextFromXML($)
{
  my $xml = shift or
    die "Must supply XML code to getFormattedTextFromXML()!\n";
  my $text = '';

  #==== Use XML::Simple to turn the XML code into a hash
  my $hr = XMLin($xml, forcearray => 0, keyattr => []);

  #===== Get key/value pairs
  $text = getKeyValuePairs($hr);

  return($text);

} # end of getFormattedTextFromXML()

sub getKeyValuePairs($)
{
  my $info_ref = shift or
    die "Must supply ref to getKeyValuePairs()!\n";
  my $str = '';
  my @keys = ();
  my $type = '';

  #===== Make sure it's a ref
  die "Not a hash ref at getKeyValuePairs()!\n" unless ref($info_ref);

  if (ref($info_ref) eq 'HASH')
  {
    @keys = keys %{$info_ref};
    $type = 'hash';
  }
  elsif (ref($info_ref) eq 'ARRAY')
  {
    @keys = @$info_ref;
    $type = 'array'; 
  }

  #===== Now walk through the keys
  for (my $i = 0; $i < scalar(@keys); $i++)
  {
    my $key = $keys[$i];
    my $value = '';

    if ($type eq 'hash')
    {
      $value = $info_ref->{$key};
    }
    elsif ($type eq 'array')
    {
      $value = ${$info_ref}[$i]; 
    }

    if (!ref($value))
    {
      $str .= "$key => $value\n";
    } # if $value is not a reference
    elsif (ref($value) eq 'SCALAR')
    {
      $str .= "$key => $$value\n";
    } # $value is a scalar ref
    elsif (ref($value) eq 'ARRAY')
    {
      $str .= getKeyValuePairs($value);
    } # if $value is an array ref
    elsif (ref($value) eq 'HASH')
    {
      if (defined($value->{value}{string}))
      {
        my $new_value = $value->{value}{string};

        if (defined($value->{name}))
        {
          my $name = $value->{name};
          $str .= "$name => $new_value\n";
          delete($value->{name});
        }
        else
        {
          $str .= "$key => $new_value\n";
        }
        delete($value->{value}{string});
      }
      elsif (defined($value->{value}{double}))
      {
        my $new_value = $value->{value}{double};
        if (defined($value->{name}))
        {
          my $name = $value->{name};
          $str .= "$name => $new_value\n";
          delete($value->{name});
        }
        else
        {
          $str .= "$key => $new_value\n";
        } 
        delete($value->{value}{double});
      }
      elsif (defined($value->{value}{int}))
      {
        my $new_value = $value->{value}{int};
        if (defined($value->{name}))
        {
          my $name = $value->{name};
          $str .= "$name => $new_value\n";
          delete($value->{name});
        }
        else
        {
          $str .= "$key => $new_value\n";
        } 
        delete($value->{value}{int});
      }
      elsif (defined($value->{name}))
      {
        my $name = $value->{name};
        $str .= "<$name>\n";
        delete($value->{name});
        $str .= getKeyValuePairs($value->{value});
        delete($value->{value});
        $str .= "</$name>\n";
      }
      elsif (defined($value->{value}{struct}{member}))
      {
        $str .= getKeyValuePairs($value->{value}{struct}{member});
        delete ($value->{value}{struct}{member});
      }
      elsif (defined($value->{value}))
      {
        delete($value->{value});
      }
      $str .= getKeyValuePairs($value);
    } # if $value is a hash ref
  } # foreach key

  return($str);

} # end of getKeyValuePairs()


#===========================================================
#    G E T  E X P E R I M E N T  I D
#
#    Parses XML-RPC call looking for the experiment ID
#
#    Arg1: Raw XML-RPC data
#    Returns: ExperimentID or undef if none found
#===========================================================
sub getExperimentID($)
{
  my $xml = shift or
    die("Must supply XML code to getExperimentID()!\n");
  my $exp_id = undef;

  #==== Use XML::Simple to turn the XML code into a hash
  my $hr = XMLin($xml, forcearray => 0, keyattr => ['name']);

  #===== Debug
  #$Data::Dumper::Indent = 1;
  #my $dump = Dumper($hr);
  #print "$dump\n";

  #===== Check extended syntax
  if (!ref($hr->{params}{param}{value}{struct}{member}{'ExperimentID'}{value}{string}) and
      $hr->{params}{param}{value}{struct}{member}{'ExperimentID'}{value}{string})
  {
    $exp_id = $hr->{params}{param}{value}{struct}{member}{'ExperimentID'}{value}{string};
  }

  #===== Check the simple command syntax
  elsif (!ref($hr->{params}{param}{value}{struct}{member}{value}{string}) and
      $hr->{params}{param}{value}{struct}{member}{value}{string})
  {
    $exp_id = $hr->{params}{param}{value}{struct}{member}{value}{string};
  }

  return($exp_id);

} # end of getExperimentID()


#===========================================================
#    G E T  C O M M A N D
#
#    Arg1: The XML text to be parsed
#===========================================================
sub getCommand($)
{
  my $xml = shift or
    die("Must supply XML code to getCommand()!\n");
  my $result = '';
  my $param_name = '';
  my $param_value = '';
       
       
  #===== Use XML::Simple to turn the XML code into a hash
  my $hr = XMLin($xml);

  #===== Start Capture
  if ($hr->{methodName} =~ m/StartExperiment/)
  {
    $result = "StartExperiment";
  } # start_capture
 
  #===== Pause Capture
  if ($hr->{methodName} =~ m/PauseExperiment/)
  {
    $result = "PauseExperiment";
  } # pause_capture
 
  #===== Stop Capture
  elsif ($hr->{methodName} =~ m/StopExperiment/)
  {
    $result = "StopExperiment";
  } # stop_capture
 
  #===== Set Capture Params
  elsif ($hr->{methodName} =~ m/ConfigureExperiment/)
  {
    $result = "ConfigureExperiment";
  } # set_capture_params

  #===== Get Experiment Status
  elsif ($hr->{methodName} =~ m/GetExperimentStatus/)
  {
    $result = "GetExperimentStatus";
  } # GetExperimentStatus

  #===== Get Experiment Settings
  elsif ($hr->{methodName} =~ m/GetExperimentSettings/)
  {
    $result = "GetExperimentSettings";
  } # GetExperimentSettings

  #===== Get device capabilities
  elsif ($hr->{methodName} =~ m/GetDeviceCapabilities/)
  {
    $result = "GetDeviceCapabilities";
  } # GetDeviceCapabilities
 
  #===== Get experiment results
  elsif ($hr->{methodName} =~ m/GetExperimentResults/)
  {
    $result = "GetExperimentResults";
  } # GetDeviceCapabilities

  #===== Chomp result
  chomp($result);
 
  return($result);
 
} # end of getCommand()

#===========================================================
#    G E T  L O C A T I O N  A N D  D E V I C E  N A M E
#
#    Parses XML-RPC call looking for the location name
#    for which we'll request device capabilities.
#
#    Arg1: Raw XML-RPC data
#    Returns: Location name
#===========================================================
sub getLocationAndDeviceName($)
{
  my $xml = shift or
    informAndDie("Must supply XML code to getLocationAndDeviceName()!\n");
  my $location_name = '';
  my $device_name = '';
                                          
  #===== Use XML::Simple to turn the XML code into a hash
  my $hr = XMLin($xml, forcearray => 0, keyattr => ['name']);

  # Debug
  #$Data::Dumper::Indent = 1;
  #my $dump = Dumper($hr);
  #print "$dump\n";

  #===== Grab the location name
  if (defined($hr->{params}{param}{value}{struct}{member}{'SiteLocation'}{value}{string}))
  {
    $location_name = $hr->{params}{param}{value}{struct}{member}{'SiteLocation'}{value}{string};
  }
  if (defined($hr->{params}{param}{value}{struct}{member}{'DeviceName'}{value}{string}))
  {
    $device_name = $hr->{params}{param}{value}{struct}{member}{'DeviceName'}{value}{string};
  }
  
  return($location_name, $device_name);
 
} # end of getLocationAndDeviceName()


#==========================================================
#    G E T  R E S P O N S E  S T R I N G
#
#    Arg1: The XML/RPC code as a scalar.
#    Returns: The response string or '' if unable to parse.
#==========================================================
sub getResponseString($)
{
  my $xml = shift or
    informAndDie("Must supply XML code to getResponseString()!\n");
  my $result = '';
 
  #===== Filter out http header stuff
  $xml =~ /(..xml\sversio.*)$/s;
  $xml = $1;

  #===== Check for faults
  my ($fault_code, $fault_string) = checkForFault($xml);
  
  #===== Handle faults
  if ($fault_string)
  {
    $result = $fault_string;
  } # if we have a fault
  else
  {
    #===== Use XML::Simple to turn the XML code into a hash
    my $xml_ref = XMLin($xml, forcearray => 0, keyattr => ['name']);

    # Debug
    #$Data::Dumper::Indent = 1;
    #return(Dumper($xml_ref));
    
    my $response = $xml_ref->{'params'}{'param'}{'value'}{'string'}; 

    $result = "$response";
  }

  return($result);

} # end of getResponseString()


#===========================================================
#    G E T  H T T P  H E A D E R
#
#    Arg1: Length of the content in bytes (zero if not sent)
#    Returns: The standard HTTP Header
#===========================================================
sub getHttpHeader(;$)
{
  my $content_length = shift;
  my $header = '';

  $content_length = '0' unless $content_length;

  $header .= "POST /RPC2 HTTP/1.0\n";
  $header .= "Content-Type: text/xml\n";
  $header .= sprintf("Content-length: %i\n\n", $content_length);

  return($header);

} # end of getHttpHeader()


1
