#! /usr/bin/perl -w

#=============================================================================
#    G I M S  A G G R E G A T E  M A N A G E R  .  P L
#
#    XML/RPC server for GIMS project
#    Written by C. Thomas
#
#    $Id: gims_aggregate_manager.cgi,v 1.18 2010/07/15 21:23:09 cthomas Exp $
#=============================================================================


#===========================================================
#    U S E  /  R E Q U I R E
#===========================================================
use CGI;                       # For form/params stuff
use XML::RPC;                  # Handles calls and responses
use XML::Simple;               # For parsing XML:RPC code
use Data::Dumper;
use Socket;                    # For 2-way communication with device
use POSIX;                     # For strftime
use GIMSDBUtils;               # To talk to database
use XmlRpcUtils;               # To create and process XML/RPC code
use strict;                    # Always a good idea


#===========================================================
#    C O N F I G
#===========================================================
my $log_file = 'gims_am.log';


#===========================================================
#    P R O T O T Y P E S
#===========================================================
sub handleInput($$);
sub openSocket($$);
sub writeToSocket($$$);
sub sendToSocket($$);
sub readFromSocket($);
sub closeSocket($);

sub getConnectInfo($$);
sub validateExpState($$$$);
sub getDeviceCapabilities($$;$;$);
sub getExperimentSettings($$;$;$);
sub formatAsReply(;$;$);
sub isDeviceCommand($);
sub getTimestamp();
sub writeToLog(;$);
sub informAndDie(;$);


#===========================================================
#    G L O B A L S
#===========================================================
my $cgi_obj = undef;
my $xml_rpc_obj = undef;

# Filehandles
my $LOG = undef;

#===========================================================
#    M A I N
#===========================================================

#==== Open log file
unless(open($LOG, ">>$log_file"))
{
  warn "Couldn't open log file! : $!\n";
}
writeToLog("started ag man\n");


#===== Get a CGI obj
$cgi_obj = new CGI;
warn "created cgi object! : $!\n";

#===== Get a new XML::RPC obj
$xml_rpc_obj = XML::RPC->new();

#===== Retrieve XML data
my $xml = $cgi_obj->param('POSTDATA');

#===== Print XML header
print $cgi_obj->header( -type => 'text/xml', -charset => 'UTF-8' );

#===== Send a response back to the client and
#      get the xml code at the same time
#      handleInput is where all the work happens!
print $xml_rpc_obj->receive($xml, \&handleInput);

#===== Close the log
close($LOG);

exit;

################## BEGIN SUBROUTINES #####################


#===========================================================
#    H A N D L E  I N P U T
#    Arg1: The XML code
#    Arg2: The XML parameters (ignored)
#===========================================================
sub handleInput($$) 
{
  my $socket = undef;
  my $raw_xml = '';
  my $response_xml = '';
  my $cmd = '';
  my $response = 'Aggregate manager generated an error.';
  my $packaged_reply = '';
  my $exp_id = '';
  my $err_str = '';

  #===== Get the xml code, store it in a global variable
  $raw_xml = shift;
  my @params = shift; # ignored

  #===== Check for XML code
  unless ($raw_xml)
  {
    $err_str = "Must supply XML code to handleInput()!\n";
    writeToLog($err_str);
    $response_xml = XmlRpcUtils::writeFaultXML(-1, $err_str);
    $packaged_reply = formatAsReply($response, $response_xml);
    return($packaged_reply);
  }

  #===== Get the ExperimentID
  $exp_id = XmlRpcUtils::getExperimentID($raw_xml);
  unless ($exp_id)
  {
    $err_str = "Unable to get ExpID from XML!\n";
    writeToLog($err_str);
    $response_xml = XmlRpcUtils::writeFaultXML(-1, $err_str);
    $packaged_reply = formatAsReply($response, $response_xml);
    return($packaged_reply);
  }

  #===== Get the command from the xml
  $cmd = XmlRpcUtils::getCommand($raw_xml);
  unless ($cmd)
  {
    $err_str = "Unable to extract command from XML!\n";
    writeToLog($err_str);
    $response_xml = XmlRpcUtils::writeFaultXML(-1, $err_str);
    $packaged_reply = formatAsReply($response, $response_xml);
    return($packaged_reply);
  }
  writeToLog("Handling command $cmd...\n");

  #===== Get a database handle
  my $dbh = GIMSDBUtils::getDBHandle();
  unless (ref($dbh))
  {
    $err_str = "Unable to get database handle!\n";
    writeToLog($err_str);
    $response_xml = XmlRpcUtils::writeFaultXML(-1, $err_str);
    $packaged_reply = formatAsReply($response, $response_xml);
    return($packaged_reply);
  }
  writeToLog("Got DB handle...\n");

  #===== Open sockets if we're talking to the device
  if (isDeviceCommand($cmd))
  {
    #===== Get host and port connection info
    my ($gims_device_host, $gims_device_port) = getConnectInfo($dbh, $exp_id);
      
    #===== If we have the hostname and port, process the command
    if ($gims_device_host and $gims_device_port)
    {
      my $valid_state = '';

      #===== Get experiment status
      my $status = GIMSDBUtils::getExpStatus($dbh, $exp_id);

      #===== Validate the state of the experiment prior to 
      #      sending the command
      writeToLog("Validating experiment state...\n");
      ($valid_state, $response_xml) = validateExpState($dbh, 
                                                       $cmd, 
                                                       $status, 
                                                       $exp_id);

      #===== Send command if state is valid to do so
      if ($valid_state)
      { 
        my $success = '';

        #===== Write code to socket
        ($success, $response_xml) = writeToSocket($gims_device_host, 
                                                  $gims_device_port, 
                                                  $raw_xml);

        #===== POST-PROCESSING
        if ($success and $cmd eq 'StartExperiment')
        {
          GIMSDBUtils::setExpStatus($dbh, $exp_id, 'running');
          writeToLog("Updated status to 'running' on ExpID: $exp_id.\n");
        }
        elsif ($success and $cmd eq 'PauseExperiment')
        {
          GIMSDBUtils::setExpStatus($dbh, $exp_id, 'paused');
          writeToLog("Updated status to 'paused' on ExpID: $exp_id.\n");
        }
        elsif ($success and $cmd eq 'StopExperiment')
        {
          GIMSDBUtils::setExpStatus($dbh, $exp_id, 'done');
          writeToLog("Updated status to 'done' on ExpID: $exp_id.\n");
        }
        elsif ($success and $cmd eq 'ConfigureExperiment')
        {
          my ($location, $device) = XmlRpcUtils::getLocationAndDeviceName($raw_xml);
          writeToLog("Updating config for $location : $device...\n");
          my $errors = GIMSDBUtils::updateExpDeviceSettings($dbh,
                                                            $exp_id, 
                                                            $location,
                                                            $device,
                                                            $raw_xml); 
          if ($errors)
          {
            writeToLog("$errors\n") if $errors;
          }
          else
          {
           writeToLog("Updated database with settings for $exp_id.\n");
          }

          #===== Update the experiment status
          GIMSDBUtils::setExpStatus($dbh, $exp_id, 'configured');
          writeToLog("Updated status to 'configured' on ExpID: $exp_id.\n");
        } # if ConfigureExperiment
        elsif ($success and $cmd eq 'GetExperimentStatus')
        {
          # FIX ME
        } # if GetExperimentStatus
      } # if state was validated
    } # if we have a hostname and port
    else
    {
      $err_str = "Couldn't get hostname and port for ExpID: $exp_id!\n";
      $response_xml = XmlRpcUtils::writeFaultXML(-1, $err_str);
    } # if we couldn't find hostname and port
  }  # if isDeviceCommand()

  ####### COMMANDS THAT DON'T TALK TO THE DEVICE BELOW #########

  elsif ($cmd eq 'GetDeviceCapabilities')
  {
    #===== Get location and device name from xml
    my ($location, $device_name) = XmlRpcUtils::getLocationAndDeviceName($raw_xml);

    $response_xml = getDeviceCapabilities($dbh,
                                          $exp_id,
                                          $location,
                                          $device_name);
  } # GetDeviceCapabilities

  elsif ($cmd eq 'GetExperimentSettings')
  {
    #===== Get location and device name from xml
    my ($location, $device_name) = XmlRpcUtils::getLocationAndDeviceName($raw_xml);

    #===== Get the settings
    $response_xml = getExperimentSettings($dbh, 
                                          $exp_id, 
                                          $location, 
                                          $device_name);
  } # GetExperimentSettings

  elsif ($cmd eq 'GetExperimentResults')
  {
    # FIX ME
  } # GetExperimentResults

  #===== Unknown command
  else
  {
    $err_str = "Unknown or unspecified command in handleInput()!\n";   
    $response_xml = XmlRpcUtils::writeFaultXML(-1, $err_str);
  } # unknown command

  #===== Package this reply as XML-RPC
  $response = "Aggregate manager received command $cmd.";
  $packaged_reply = formatAsReply($response, $response_xml);
 
  #===== Debug
  #writeToLog("Response XML:\n");
  #writeToLog($response_xml . "\n");
  #writeToLog("Packaged reply:\n");
  #writeToLog($packaged_reply);
   
  return($packaged_reply);
  
} # handleInput()


#===========================================================
#    O P E N  S O C K E T
#
#    Arg1: The hostname.
#    Arg2: The port.
#    Returns: The socket filehandle and the socket.
#===========================================================
sub openSocket($$)
{  
  my $host = shift or
    informAndDie("Must supply hostname to openSocket()\n");
  my $port = shift or
    informAndDie("Must supply port number to openSocket()\n");
  my $SOCK_FH = undef;
  my $sock = undef;
  
  my $proto = getprotobyname('tcp');
  my $iaddr = inet_aton($host);
  my $paddr = sockaddr_in($port, $iaddr);
 
  #===== Create the socket and store it in the filehandle 
  writeToLog("Calling socket()...\n");
  unless(socket($SOCK_FH, PF_INET, SOCK_STREAM, $proto))
  {
    writeToLog("ERROR: socket: $!\n");
    return(undef, undef);
  }

  #===== Set filehandle to unbuffered mode
  writeToLog("Setting filehandle to unbuffered mode...\n");
  my $old_fh = select($SOCK_FH);
  $| = 1;
  $sock = select($old_fh);
	
  #==== Connect socket to our server address 
  writeToLog("Connecting...\n");
  unless(connect($SOCK_FH, $paddr))
  {
    writeToLog("ERROR: connect() failed : $!\n");
    return(undef, undef);
  }
  print "Client connected \n";
  
  writeToLog("Socket opened to ${host}\:${port}.\n");
    
  return($SOCK_FH, $sock);

} # end of openSocket()


#===========================================================
#    W R I T E  T O  S O C K E T
#
#    Arg1: The hostname
#    Arg2: The port
#    Returns: The response XML
#===========================================================
sub writeToSocket($$$)
{
  my $host = shift or
    die "Must supplly hostname to writeToSocket()!\n";
  my $port = shift or
    die "Must supply port to writeToSocket()!\n";
  my $raw_xml = shift or
    die "Must supply XML/RPC code to writeToSocket()!\n";
  my $SOCK_FH = undef;
  my $socket = undef;
  my $response_xml = '';
  my $success = 0;

  #===== Add HTTP header to commands going to the device
  my $xml_header = XmlRpcUtils::getHttpHeader(length($raw_xml));
  my $xml_plus_header = $xml_header . $raw_xml;

  #===== Open a socket to the capture device
  writeToLog("Opening socket for $host, $port...\n");
  ($SOCK_FH, $socket) = openSocket($host, $port); 

  #===== Send the data
  if ($socket)
  {
    writeToLog("Sending to socket...\n");
    sendToSocket($xml_plus_header, $SOCK_FH);

    #===== Listen for XML-RPC response
    writeToLog("Listening for response...\n");
    $response_xml = readFromSocket($SOCK_FH);
		
    #===== Close the socket
    writeToLog("Closing socket...\n");
    closeSocket($SOCK_FH);

    #===== Set the flag
    $success = 1;
  } # if we have a socket
  else
  {
    writeToLog("No socket...\n");
    my $err_str = "Couldn't get socket for $host:$port!\n"; 
    $err_str .= "Check that device is ready to receive incoming commands.\n"; 
    $response_xml = XmlRpcUtils::writeFaultXML(-1, $err_str);
  } # no socket

  return($success, $response_xml);

} # writeToSocket()


#===========================================================
#    S E N D  T O  S O C K E T
#
#    Arg1: The message to send
#    Arg2: The socket filehandle.
#===========================================================
sub sendToSocket($$)
{
  my $msg = shift or
    informAndDie("Must supply msg to sendToSocket()!\n");
  my $socket = shift or
    informAndDie("Must supply socket filehandle to sendToSocket()!\n");
   
  print $socket "$msg\n\n"; # blank like is important!

  # Debug
  #writeToLog("CLIENT:\n$msg");
   
  return;
  
} # end of sendToSocket()


#===========================================================
#    R E A D  F R O M  S O C K E T
#
#    Arg1: The socket filehandle.
#    Returns: The message read from the socket.
#===========================================================
sub readFromSocket($)
{
  my $socket_filehandle = shift
    or informAndDie("Must supply filehandle to readFromSocket()!\n");
  my $msg = '';
	
  while (my $line = <$socket_filehandle>)
  {
    if ($line)
    {
      # debug
      #writeToLog("DEVICE: $line");
      $msg .= $line;
    }
  }
        
  return($msg);
  
} # end of readFromSocket()


#===========================================================
#    C L O S E  S O C K E T
#
#    Arg1: The socket filehandle.
#===========================================================
sub closeSocket($)
{  
  my $socket_filehandle = shift or
    informAndDie("Must supply filehandle to closeSocket()!\n");
  close ($socket_filehandle) or 
    informAndDie("close: $!");

  writeToLog("Socket closed.\n");

  return;

} # end of closeSocket()


#===========================================================
#    G E T  C O N N E C T  I N F O
#
#    Gets the hostname and port for the experiment in 
#    question.
#
#    Arg1: A database handle
#    Arg2: The ExperimentID
#    Returns: The hostname and port for the given exp_id
#===========================================================
sub getConnectInfo($$)
{
  my $dbh = shift or
    informAndDie("Must supply database handle to getConnectInfo()!\n");
  my $exp_id = shift or
    informAndDie("Must supply ExperimentID to getConnectInfo()!\n");
  my %devices = ();
  my $hostname = '';
  my $port = '';

  #===== Get information about the capture device
  my $errors = GIMSDBUtils::getExpDevices($dbh, 
                                          \%devices, 
                                          $exp_id, 
                                          'PKT_CAP');
  #===== Should be only one device
  my $num_devices = scalar(keys %{$devices{$exp_id}});
  if ($num_devices != 1)
  {
    writeToLog("Warning: $num_devices devices returned for ExperimentID=$exp_id\n");   
  }

  #===== Warn about DB errors, don't die
  if ($errors)
  {
    writeToLog("ERROR: $errors\n");
  }

  #===== Get the hostname and port
  foreach my $location (keys %{$devices{$exp_id}})
  {
    foreach my $device_name (keys %{$devices{$exp_id}{$location}})
    {
      $hostname = $devices{$exp_id}{$location}{$device_name}{'hostname'};
      $port = $devices{$exp_id}{$location}{$device_name}{'port'};
    } # foreach device_name
  } # foreach location

  return($hostname, $port);

} # end of getConnectInfo()


#===========================================================
#    V A L I D A T E  E X P  S T A T E
#
#    Arg1: The database handle.
#    Arg2: The command.
#    Arg3: The experiment state
#    Arg4: The ExperimentID
#    Returns: An array with two elements
#             0) Success or failure (boolean)
#             1) Any response XML (in case of failure)
#===========================================================
sub validateExpState($$$$)
{
  my $dbh = shift or
    informAndDie("Must supply database handle to validateExpState()!\n");
  my $cmd = shift or
    informAndDie("Must supply command to validateExpState()!\n");
  my $status = shift or
    informAndDie("Must supply status to validateExpState()!\n");
  my $exp_id = shift or
    informAndDie("Must supply ExperimentID to validateExpState()!\n");
  my $response_xml = '';
  my $success = 0;
  my $err_str = '';

  #===== Now make sure this command can be executed when the experiment
  #      is in the state given.
  if ($cmd eq 'StartExperiment')
  {
    if (!defined($status) or $status eq 'instantiated')
    {
      $err_str = "Experiment $exp_id is not configured!"; 
      $success = 0;
      $response_xml = XmlRpcUtils::writeFaultXML(-1, $err_str); 
    } # instantiated
    elsif ($status eq 'configured')
    {
      $success = 1;
    } # configured
    elsif ($status eq 'running')
    {
      $err_str = "Experiment $exp_id is already running!";
      $success = 0;
      $response_xml = XmlRpcUtils::writeFaultXML(-1, $err_str); 
    } # running
    elsif ($status eq 'paused')
    {
      $success = 1;
    } # paused
    elsif ($status eq 'done')
    {
      $err_str = "Cannot restart finished experiment $exp_id?";
      $success = 0;
      $response_xml = XmlRpcUtils::writeFaultXML(-1, $err_str); 
    } # done
  } # StartExperiment

  #===== PauseExperiment
  elsif ($cmd eq 'PauseExperiment')
  {
    if (!defined($status) or $status eq 'instantiated')
    {
      $err_str = "Experiment $exp_id is not configured!"; 
      $success = 0;
      $response_xml = XmlRpcUtils::writeFaultXML(-1, $err_str); 
    } # instantiated
    elsif ($status eq 'configured')
    {
      $err_str = "Experiment $exp_id is not running!";
      $success = 0;
      $response_xml = XmlRpcUtils::writeFaultXML(-1, $err_str); 
    } # configured
    elsif ($status eq 'running')
    {
      $success = 1;
    } # running
    elsif ($status eq 'paused')
    {
      $err_str = "Experiment $exp_id is already paused!";
      $success = 0;
      $response_xml = XmlRpcUtils::writeFaultXML(-1, $err_str); 
    } # paused
    elsif ($status eq 'done')
    {
      $err_str = "Cannot pause finished experiment $exp_id!";
      $success = 0;
      $response_xml = XmlRpcUtils::writeFaultXML(-1, $err_str); 
    } # done
  } # PauseExperiment

  #===== StopExperiment
  elsif ($cmd eq 'StopExperiment')
  {
    if (!defined($status) or $status eq 'instantiated')
    {
      $err_str = "Experiment $exp_id is not configured!"; 
      $success = 0;
      $response_xml = XmlRpcUtils::writeFaultXML(-1, $err_str); 
    } # instantiated
    elsif ($status eq 'configured')
    {
      $err_str = "Experiment $exp_id is not running!";
      $success = 0;
      $response_xml = XmlRpcUtils::writeFaultXML(-1, $err_str); 
    } # configured
    elsif ($status eq 'running')
    {
      $success = 1;
    } # running
    elsif ($status eq 'paused')
    {
      $success = 1;
    } # paused
    elsif ($status eq 'done')
    {
      $err_str = "Experiment $exp_id has already been terminated!";
      $success = 0;
      $response_xml = XmlRpcUtils::writeFaultXML(-1, $err_str); 
    } # done
  } # StopExperiment

  #===== ConfigureExperiment
  elsif ($cmd eq 'ConfigureExperiment')
  {
    if (!defined($status) or $status eq 'instantiated')
    {
      $success = 1;
    } # instantiated
    elsif ($status eq 'configured')
    {
      $success = 1;
    } # configured
    elsif ($status eq 'running')
    {
      $err_str = "Cannot configure running experiment $exp_id!";
      $success = 0;
      $response_xml = XmlRpcUtils::writeFaultXML(-1, $err_str); 
    } # running
    elsif ($status eq 'paused')
    {
      $err_str = "Cannot configure paused experiment $exp_id!";
      $success = 0;
      $response_xml = XmlRpcUtils::writeFaultXML(-1, $err_str); 
    } # paused
    elsif ($status eq 'done')
    {
      $err_str = "Cannot configure finished experiment $exp_id!";
      $success = 0;
      $response_xml = XmlRpcUtils::writeFaultXML(-1, $err_str); 
    } # done
  } # ConfigureExperiment

  #===== GetExperimentStatus
  elsif ($cmd eq 'GetExperimentStatus')
  {
    if (!defined($status))
    {
      $err_str = "Experiment $exp_id not found in database!";
      $success = 0;
      $response_xml = XmlRpcUtils::writeFaultXML(-1, $err_str); 
    }
    elsif ($status eq 'instantiated')
    {
      $err_str = "Experiment $exp_id is not running!";
      $success = 0;
      $response_xml = XmlRpcUtils::writeFaultXML(-1, $err_str); 
    } # instantiated
    elsif ($status eq 'configured')
    {
      $err_str = "Experiment $exp_id is not running!";
      $success = 0;
      $response_xml = XmlRpcUtils::writeFaultXML(-1, $err_str); 
    } # configured
    elsif ($status eq 'running')
    {
      $success = 1;
    } # running
    elsif ($status eq 'paused')
    {
      $success = 1;
    } # paused
    elsif ($status eq 'done')
    {
      $err_str = "Experiment $exp_id is done!";
      $success = 0;
      $response_xml = XmlRpcUtils::writeFaultXML(-1, $err_str); 
    } # done
  } # GetExperimentStatus

  #===== GetExperimentResults
  elsif ($cmd eq 'GetExperimentResults')
  {
    if (!defined($status))
    {
      $err_str = "Experiment $exp_id not found in database!";
      $success = 0;
      $response_xml = XmlRpcUtils::writeFaultXML(-1, $err_str); 
    }
    elsif ($status eq 'instantiated')
    {
      $err_str = "Experiment $exp_id not started yet!"; 
      $success = 0;
      $response_xml = XmlRpcUtils::writeFaultXML(-1, $err_str); 
    } # instantiated
    elsif ($status eq 'configured')
    {
      $err_str = "Experiment $exp_id not started yet!";
      $success = 0;
      $response_xml = XmlRpcUtils::writeFaultXML(-1, $err_str); 
    } # configured
    elsif ($status eq 'running')
    {
      $err_str = "Experiment $exp_id is still running!";
      $success = 0;
      $response_xml = XmlRpcUtils::writeFaultXML(-1, $err_str); 
    } # running
    elsif ($status eq 'paused')
    {
      $err_str = "Must stop $exp_id before results will be available!";
      $success = 0;
      $response_xml = XmlRpcUtils::writeFaultXML(-1, $err_str); 
    } # paused
    elsif ($status eq 'done')
    {
      $success = 1;
    } # done
  } # GetExperimentResults

  return($success, $response_xml);

} # validateExpState()


#===========================================================
#    G E T  D E V I C E  C A P A B I L I T I E S
#
#    Arg1: Database handle
#    Arg2: ExperimentID
#    Arg3: Location (optional)
#    Arg4: Device name (optional)
#    Returns: XML/RPC with caps or error, as appropriate
#===========================================================
sub getDeviceCapabilities($$;$;$)
{
  my $dbh = shift or
    informAndDie("Must supply database handle to getExperimentSettings()!\n");
  my $exp_id = shift or
    informAndDie("Must supply ExpID to getExperimentSettings()!\n");
  my $location = shift;
  my $device_name = shift;
  my $response_xml = '';

  #===== Get the capabilities
  if ($location and $device_name)
  {
    #===== Get the device capabilities from the database
    writeToLog("Getting capabilities for $location : $device_name...\n");
    $response_xml = GIMSDBUtils::getDeviceCapabilities($dbh,
                                                       $location, 
                                                       $device_name);
  } # if we have location and device_name
  else
  {
    my $err_str = "Missing location or device_name in handleInput()!\n";
    $response_xml = XmlRpcUtils::writeFaultXML(-1, $err_str);
  } # no location_name or device_name

  return($response_xml);

} # end of getDeviceCapabilities()


#===========================================================
#    G E T  E X P E R I M E N T  S E T T I N G S
#
#    Arg1: Database handle
#    Arg2: ExperimentID
#    Arg3: Location (optional)
#    Arg4: Device name (optional)
#    Returns: XML/RPC with settings or error, as appropriate
#===========================================================
sub getExperimentSettings($$;$;$)
{
  my $dbh = shift or
    informAndDie("Must supply database handle to getExperimentSettings()!\n");
  my $exp_id = shift or
    informAndDie("Must supply ExpID to getExperimentSettings()!\n");
  my $location = shift;
  my $device_name = shift;
  my $response_xml = '';
  my %settings = ();

  #===== Get the device capabilities from the database
  writeToLog("Getting settings for ExpID: $exp_id...\n");
  GIMSDBUtils::getExpDeviceInfo($dbh,
                                \%settings, 
                                $exp_id,
                                $location,
                                $device_name);

  #===== Pull out each device's settings
  foreach my $location (keys %{$settings{$exp_id}})
  {
    foreach my $device (keys %{$settings{$exp_id}{$location}})
    {
      if ($settings{$exp_id}{$location}{$device}{'settings'})
      {
        $response_xml .= $settings{$exp_id}{$location}{$device}{'settings'}; 
      } # if we have settings
    } # for each device
  } # for each location

  #===== Send a fault if no settings found
  unless ($response_xml)
  {
    my $err_str = "No exp settings for";
    $err_str .= " ExpID: $exp_id";
    $err_str .= " Location: $location" if $location;
    $err_str .= " Device: $device_name" if $device_name;
    $err_str .= "?";
    $response_xml = XmlRpcUtils::writeFaultXML(-1, $err_str);
  } # if we have no response

  return($response_xml);

} # end of getExperimentSettings()


#===========================================================
#    F O R M A T  A S  R E P L Y
#
#    Arg1: Data from the aggregate manager we wish to return
#    Arg2: Data from the device we wish to return
#===========================================================
sub formatAsReply(;$;$)
{
  my $agr_mgr_data = shift;
  my $device_response_data = shift;
  my $reply = '';
  
  writeToLog("Formulating reply...\n");
  
  #===== Encapsulate the method name in an XML-RPC response
  $reply .= '<?xml version="1.0" encoding="UTF-8" ?>' . "\n";
  $reply .= "<methodResponse>\n";
  $reply .= "<params><param><value><struct>\n";
  
  #===== Data from the Aggregate Manager we wish to return to user
  $reply .= "<member>\n";
  $reply .= "<name>agr_mgr_data</name>\n";
  $reply .= "<value><string><![CDATA[$agr_mgr_data]]></string></value>\n";
  $reply .= "</member>\n";
  
  #===== Response from the GIMS device
  $reply .= "<member>\n";
  $reply .= "<name>device_response_data</name>\n";
  $reply .= "<value><string><![CDATA[$device_response_data]]></string></value>\n";
  $reply .= "</member>\n";
  
  $reply .= "</struct></value></param></params>\n";
  $reply .= "</methodResponse>\n";
    
  return($reply);
  
} # end of formatAsReply()


#===========================================================
#    I S  D E V I C E  C O M M A N D
#
#    Arg1: The command.
#    Returns: 1 if the command involves the device.
#             0 otherwise.
#===========================================================
sub isDeviceCommand($)
{
  my $cmd = shift or
    informAndDie("Must supply command to isDeviceCommand()!\n");
  my $device_command = 0;
  if ( $cmd eq 'ConfigureExperiment' or 
       $cmd eq 'StartExperiment' or 
       $cmd eq 'PauseExperiment' or 
       $cmd eq 'StopExperiment' or
       $cmd eq 'GetExperimentStatus' or
       $cmd eq 'GetExperimentResults' )
  {
    $device_command = 1;
  }

  return($device_command);

} # end of isDeviceCommand()

#===========================================================
#    G E T  T I M E S T A M P
#===========================================================
sub getTimestamp()
{
  return(strftime("%Y-%m-%d %H:%M:%S", localtime));
} # end of getTimestamp()


#===========================================================
#    W R I T E  T O  L O G
#===========================================================
sub writeToLog(;$)
{
  my $output_str = shift;

  if ($output_str)
  {
    print $LOG getTimestamp() . ": ";
    print $LOG $output_str;
  }
  
  return;
  
} # end of writeToLog()


#===========================================================
#    I N F O R M  A N D  D I E
#    Tries to write to log, then dies
#    Arg1: The error message
#===========================================================
sub informAndDie(;$)
{
  my $output_str = shift;

  if ($output_str)
  {
    print $LOG getTimestamp() . ": ";
    print $LOG $output_str;
  }
  
  die "$output_str\n";
    
} # end of informAndDie()
